#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for Component objects

Author: Andrew Lee
"""
import pytest
import re
import types

from pyomo.environ import ConcreteModel, Set, Param, Var, units as pyunits

from idaes.core.base.components import (
    Component,
    Solute,
    Solvent,
    Ion,
    Anion,
    Cation,
    Apparent,
)
from idaes.core.base.phases import (
    LiquidPhase,
    VaporPhase,
    SolidPhase,
    Phase,
    PhaseType,
    AqueousPhase,
)
from idaes.core.util.exceptions import ConfigurationError, PropertyPackageError
from idaes.core.base.property_meta import PropertyClassMetadata, UnitSet


class TestComponent:
    @pytest.fixture(scope="class")
    def m(self):
        m = ConcreteModel()

        m.meta_object = PropertyClassMetadata()
        m.meta_object._default_units.set_units(
            temperature=pyunits.K,
            mass=pyunits.kg,
            length=pyunits.m,
            time=pyunits.s,
            amount=pyunits.mol,
        )

        def get_metadata(self):
            return m.meta_object

        m.get_metadata = types.MethodType(get_metadata, m)

        m.comp = Component()
        m.comp2 = Component()

        return m

    @pytest.mark.unit
    def test_config(self, m):
        assert m.comp.config.valid_phase_types is None
        assert m.comp.config.elemental_composition is None
        assert not m.comp.config._component_list_exists
        assert m.comp.config.henry_component is None
        assert m.comp.config.has_vapor_pressure

    @pytest.mark.unit
    def test_populate_component_list(self, m):
        assert isinstance(m.component_list, Set)

        for j in m.component_list:
            assert j in ["comp", "comp2"]

    @pytest.mark.unit
    def test_is_solute(self, m):
        with pytest.raises(
            TypeError,
            match=re.escape(
                "comp Generic Component objects do not "
                "support is_solute() method. Use a Solvent or "
                "Solute Component instead."
            ),
        ):
            m.comp.is_solute()

    @pytest.mark.unit
    def test_is_solvent(self, m):
        with pytest.raises(
            TypeError,
            match=re.escape(
                "comp Generic Component objects do not "
                "support is_solvent() method. Use a Solvent or "
                "Solute Component instead."
            ),
        ):
            m.comp.is_solvent()

    @pytest.mark.unit
    def test_is_phase_valid_no_assignment(self, m):
        with pytest.raises(TypeError):
            m.comp._is_phase_valid("foo")

    @pytest.mark.unit
    def test_is_phase_valid_liquid(self, m):
        m.comp3 = Component(valid_phase_types=PhaseType.liquidPhase)

        m.Liq = LiquidPhase()
        m.Sol = SolidPhase()
        m.Vap = VaporPhase()
        m.Aqu = AqueousPhase()
        m.Phase = Phase()

        assert m.comp3._is_phase_valid(m.Liq)
        assert not m.comp3._is_phase_valid(m.Sol)
        assert not m.comp3._is_phase_valid(m.Vap)
        assert not m.comp3._is_phase_valid(m.Aqu)
        assert not m.comp3._is_phase_valid(m.Phase)

    @pytest.mark.unit
    def test_is_phase_valid_vapor(self, m):
        m.comp4 = Component(valid_phase_types=PhaseType.vaporPhase)

        assert not m.comp4._is_phase_valid(m.Liq)
        assert not m.comp4._is_phase_valid(m.Sol)
        assert m.comp4._is_phase_valid(m.Vap)
        assert not m.comp4._is_phase_valid(m.Aqu)
        assert not m.comp4._is_phase_valid(m.Phase)

    @pytest.mark.unit
    def test_is_phase_valid_solid(self, m):
        m.comp5 = Component(valid_phase_types=PhaseType.solidPhase)

        assert not m.comp5._is_phase_valid(m.Liq)
        assert m.comp5._is_phase_valid(m.Sol)
        assert not m.comp5._is_phase_valid(m.Vap)
        assert not m.comp5._is_phase_valid(m.Aqu)
        assert not m.comp5._is_phase_valid(m.Phase)

    @pytest.mark.unit
    def test_is_phase_valid_aqueous(self, m):
        m.comp6 = Component(valid_phase_types=PhaseType.aqueousPhase)

        assert not m.comp6._is_phase_valid(m.Liq)
        assert not m.comp6._is_phase_valid(m.Sol)
        assert not m.comp6._is_phase_valid(m.Vap)
        # Generic components are never valid in the aqueous phase
        assert not m.comp6._is_phase_valid(m.Aqu)
        assert not m.comp6._is_phase_valid(m.Phase)

    @pytest.mark.unit
    def test_is_phase_valid_LV(self, m):
        m.comp7 = Component(
            valid_phase_types=[PhaseType.liquidPhase, PhaseType.vaporPhase]
        )

        assert m.comp7._is_phase_valid(m.Liq)
        assert not m.comp7._is_phase_valid(m.Sol)
        assert m.comp7._is_phase_valid(m.Vap)
        assert not m.comp7._is_phase_valid(m.Aqu)
        assert not m.comp7._is_phase_valid(m.Phase)

    @pytest.mark.unit
    def test_create_parameters(self):
        m = ConcreteModel()

        m.meta_object = PropertyClassMetadata()

        def get_metadata(self):
            return m.meta_object

        m.get_metadata = types.MethodType(get_metadata, m)
        m.meta_object._default_units.set_units(
            temperature=pyunits.K,
            mass=pyunits.kg,
            length=pyunits.m,
            time=pyunits.s,
            amount=pyunits.mol,
        )

        m.comp = Component(
            parameter_data={
                "mw": 10,
                "compress_fact_crit": 1,
                "dens_mol_crit": 55,
                "pressure_crit": 1e5,
                "temperature_crit": 500,
            }
        )

        assert isinstance(m.comp.mw, Param)
        assert m.comp.mw.value == 10

        assert isinstance(m.comp.compress_fact_crit, Var)
        assert m.comp.compress_fact_crit.value == 1

        assert isinstance(m.comp.dens_mol_crit, Var)
        assert m.comp.dens_mol_crit.value == 55

        assert isinstance(m.comp.pressure_crit, Var)
        assert m.comp.pressure_crit.value == 1e5

        assert isinstance(m.comp.temperature_crit, Var)
        assert m.comp.temperature_crit.value == 500

    @pytest.mark.unit
    def test_create_parameters_convert(self):
        m = ConcreteModel()

        m.meta_object = PropertyClassMetadata()

        def get_metadata(self):
            return m.meta_object

        m.get_metadata = types.MethodType(get_metadata, m)
        m.meta_object._default_units.set_units(
            temperature=pyunits.K,
            mass=pyunits.kg,
            length=pyunits.m,
            time=pyunits.s,
            amount=pyunits.mol,
        )

        m.comp = Component(
            parameter_data={
                "mw": (10, pyunits.g / pyunits.mol),
                "pressure_crit": (1, pyunits.bar),
                "temperature_crit": (900, pyunits.degR),
            }
        )

        assert isinstance(m.comp.mw, Param)
        assert m.comp.mw.value == 1e-2

        assert isinstance(m.comp.pressure_crit, Var)
        assert m.comp.pressure_crit.value == 1e5

        assert isinstance(m.comp.temperature_crit, Var)
        assert m.comp.temperature_crit.value == 500


class TestSolute:
    @pytest.fixture(scope="class")
    def m(self):
        m = ConcreteModel()

        m.meta_object = PropertyClassMetadata()
        m.meta_object._default_units.set_units(
            temperature=pyunits.K,
            mass=pyunits.kg,
            length=pyunits.m,
            time=pyunits.s,
            amount=pyunits.mol,
        )

        def get_metadata(self):
            return m.meta_object

        m.get_metadata = types.MethodType(get_metadata, m)

        m.comp = Solute()

        return m

    @pytest.mark.unit
    def test_config(self, m):
        assert m.comp.config.valid_phase_types is None
        assert not m.comp.config._component_list_exists

    @pytest.mark.unit
    def test_populate_component_list(self, m):
        assert isinstance(m.component_list, Set)

        for j in m.component_list:
            assert j in ["comp"]

        assert isinstance(m.solute_set, Set)

        for j in m.solute_set:
            assert j in ["comp"]

    @pytest.mark.unit
    def test_is_solute(self, m):
        assert m.comp.is_solute()

    @pytest.mark.unit
    def test_is_solvent(self, m):
        assert not m.comp.is_solvent()

    @pytest.mark.unit
    def test_is_phase_valid_no_assignment(self, m):
        with pytest.raises(TypeError):
            m.comp._is_phase_valid("foo")

    @pytest.mark.unit
    def test_is_phase_valid_liquid(self, m):
        m.comp3 = Solute(valid_phase_types=PhaseType.liquidPhase)

        m.Liq = LiquidPhase()
        m.Sol = SolidPhase()
        m.Vap = VaporPhase()
        m.Aqu = AqueousPhase()
        m.Phase = Phase()

        assert m.comp3._is_phase_valid(m.Liq)
        assert not m.comp3._is_phase_valid(m.Sol)
        assert not m.comp3._is_phase_valid(m.Vap)
        assert not m.comp3._is_phase_valid(m.Aqu)
        assert not m.comp3._is_phase_valid(m.Phase)

    @pytest.mark.unit
    def test_is_phase_valid_vapor(self, m):
        m.comp4 = Solute(valid_phase_types=PhaseType.vaporPhase)

        assert not m.comp4._is_phase_valid(m.Liq)
        assert not m.comp4._is_phase_valid(m.Sol)
        assert m.comp4._is_phase_valid(m.Vap)
        assert not m.comp4._is_phase_valid(m.Aqu)
        assert not m.comp4._is_phase_valid(m.Phase)

    @pytest.mark.unit
    def test_is_phase_valid_solid(self, m):
        m.comp5 = Solute(valid_phase_types=PhaseType.solidPhase)

        assert not m.comp5._is_phase_valid(m.Liq)
        assert m.comp5._is_phase_valid(m.Sol)
        assert not m.comp5._is_phase_valid(m.Vap)
        assert not m.comp5._is_phase_valid(m.Aqu)
        assert not m.comp5._is_phase_valid(m.Phase)

    @pytest.mark.unit
    def test_is_phase_valid_aqueous(self, m):
        m.comp6 = Solute(valid_phase_types=PhaseType.aqueousPhase)

        assert not m.comp6._is_phase_valid(m.Liq)
        assert not m.comp6._is_phase_valid(m.Sol)
        assert not m.comp6._is_phase_valid(m.Vap)
        assert m.comp6._is_phase_valid(m.Aqu)
        assert not m.comp6._is_phase_valid(m.Phase)

    @pytest.mark.unit
    def test_is_phase_valid_LV(self, m):
        m.comp7 = Solute(
            valid_phase_types=[PhaseType.liquidPhase, PhaseType.vaporPhase]
        )

        assert m.comp7._is_phase_valid(m.Liq)
        assert not m.comp7._is_phase_valid(m.Sol)
        assert m.comp7._is_phase_valid(m.Vap)
        assert not m.comp7._is_phase_valid(m.Aqu)
        assert not m.comp7._is_phase_valid(m.Phase)


class TestSovent:
    @pytest.fixture(scope="class")
    def m(self):
        m = ConcreteModel()

        m.meta_object = PropertyClassMetadata()
        m.meta_object._default_units.set_units(
            temperature=pyunits.K,
            mass=pyunits.kg,
            length=pyunits.m,
            time=pyunits.s,
            amount=pyunits.mol,
        )

        def get_metadata(self):
            return m.meta_object

        m.get_metadata = types.MethodType(get_metadata, m)

        m.comp = Solvent()

        return m

    @pytest.mark.unit
    def test_config(self, m):
        assert m.comp.config.valid_phase_types is None
        assert not m.comp.config._component_list_exists

    @pytest.mark.unit
    def test_populate_component_list(self, m):
        assert isinstance(m.component_list, Set)

        for j in m.component_list:
            assert j in ["comp"]

        assert isinstance(m.solvent_set, Set)

        for j in m.solvent_set:
            assert j in ["comp"]

    @pytest.mark.unit
    def test_is_solute(self, m):
        assert not m.comp.is_solute()

    @pytest.mark.unit
    def test_is_solvent(self, m):
        assert m.comp.is_solvent()

    @pytest.mark.unit
    def test_is_phase_valid_no_assignment(self, m):
        with pytest.raises(TypeError):
            m.comp._is_phase_valid("foo")

    @pytest.mark.unit
    def test_is_phase_valid_liquid(self, m):
        m.comp3 = Solvent(valid_phase_types=PhaseType.liquidPhase)

        m.Liq = LiquidPhase()
        m.Sol = SolidPhase()
        m.Vap = VaporPhase()
        m.Aqu = AqueousPhase()
        m.Phase = Phase()

        assert m.comp3._is_phase_valid(m.Liq)
        assert not m.comp3._is_phase_valid(m.Sol)
        assert not m.comp3._is_phase_valid(m.Vap)
        assert not m.comp3._is_phase_valid(m.Aqu)
        assert not m.comp3._is_phase_valid(m.Phase)

    @pytest.mark.unit
    def test_is_phase_valid_vapor(self, m):
        m.comp4 = Solvent(valid_phase_types=PhaseType.vaporPhase)

        assert not m.comp4._is_phase_valid(m.Liq)
        assert not m.comp4._is_phase_valid(m.Sol)
        assert m.comp4._is_phase_valid(m.Vap)
        assert not m.comp4._is_phase_valid(m.Aqu)
        assert not m.comp4._is_phase_valid(m.Phase)

    @pytest.mark.unit
    def test_is_phase_valid_solid(self, m):
        m.comp5 = Solvent(valid_phase_types=PhaseType.solidPhase)

        assert not m.comp5._is_phase_valid(m.Liq)
        assert m.comp5._is_phase_valid(m.Sol)
        assert not m.comp5._is_phase_valid(m.Vap)
        assert not m.comp5._is_phase_valid(m.Aqu)
        assert not m.comp5._is_phase_valid(m.Phase)

    @pytest.mark.unit
    def test_is_phase_valid_LV(self, m):
        m.comp6 = Solvent(
            valid_phase_types=[PhaseType.liquidPhase, PhaseType.vaporPhase]
        )

        assert m.comp6._is_phase_valid(m.Liq)
        assert not m.comp6._is_phase_valid(m.Sol)
        assert m.comp6._is_phase_valid(m.Vap)
        assert not m.comp6._is_phase_valid(m.Aqu)
        assert not m.comp6._is_phase_valid(m.Phase)

    @pytest.mark.unit
    def test_is_phase_valid_aqueous(self, m):
        m.comp7 = Solvent(valid_phase_types=PhaseType.aqueousPhase)

        assert not m.comp7._is_phase_valid(m.Liq)
        assert not m.comp7._is_phase_valid(m.Sol)
        assert not m.comp7._is_phase_valid(m.Vap)
        assert m.comp7._is_phase_valid(m.Aqu)
        assert not m.comp7._is_phase_valid(m.Phase)


class TestIon:
    @pytest.mark.unit
    def test_electrolyte(self):
        m = ConcreteModel()

        m.meta_object = PropertyClassMetadata()

        def get_metadata(self):
            return m.meta_object

        m.get_metadata = types.MethodType(get_metadata, m)

        with pytest.raises(
            NotImplementedError,
            match="comp The IonData component class is intended as a base "
            "class for the AnionData and CationData classes, and should "
            "not be used directly",
        ):
            m.comp = Ion(_electrolyte=True)

    @pytest.mark.unit
    def test_not_electrolyte(self):
        m = ConcreteModel()

        m.meta_object = PropertyClassMetadata()
        m.meta_object._default_units.set_units(
            temperature=pyunits.K,
            mass=pyunits.kg,
            length=pyunits.m,
            time=pyunits.s,
            amount=pyunits.mol,
        )

        def get_metadata(self):
            return m.meta_object

        m.get_metadata = types.MethodType(get_metadata, m)

        with pytest.raises(
            PropertyPackageError,
            match="comp Ion Component types should only be used with Aqueous Phases",
        ):
            m.comp = Ion(_electrolyte=False)


class TestAnion:
    @pytest.mark.unit
    def test_not_electrolyte(self):
        m = ConcreteModel()

        m.meta_object = PropertyClassMetadata()
        m.meta_object._default_units.set_units(
            temperature=pyunits.K,
            mass=pyunits.kg,
            length=pyunits.m,
            time=pyunits.s,
            amount=pyunits.mol,
        )

        def get_metadata(self):
            return m.meta_object

        m.get_metadata = types.MethodType(get_metadata, m)

        with pytest.raises(
            PropertyPackageError,
            match="comp Ion Component types should only be used with Aqueous Phases",
        ):
            m.comp = Anion(_electrolyte=False)

    @pytest.fixture(scope="class")
    def m(self):
        m = ConcreteModel()

        m.meta_object = PropertyClassMetadata()
        m.meta_object._default_units.set_units(
            temperature=pyunits.K,
            mass=pyunits.kg,
            length=pyunits.m,
            time=pyunits.s,
            amount=pyunits.mol,
        )

        def get_metadata(self):
            return m.meta_object

        m.get_metadata = types.MethodType(get_metadata, m)

        m.anion_set = Set()

        m.comp = Anion(charge=-1, _electrolyte=True)

        return m

    @pytest.mark.unit
    def test_config(self, m):
        assert "valid_phase_types" not in m.comp.config
        assert m.comp.config.charge == -1
        assert not m.comp.config._component_list_exists
        assert not m.comp.config.has_vapor_pressure
        with pytest.raises(ValueError):
            m.comp.config.has_vapor_pressure = True

    @pytest.mark.unit
    def test_populate_component_list(self, m):
        assert isinstance(m.anion_set, Set)

        for j in m.anion_set:
            assert j in ["comp"]

    @pytest.mark.unit
    def test_is_solute(self, m):
        assert m.comp.is_solute()

    @pytest.mark.unit
    def test_is_solvent(self, m):
        assert not m.comp.is_solvent()

    @pytest.mark.unit
    def test_invalid_charge(self, m):
        with pytest.raises(
            ConfigurationError,
            match="an received invalid value for charge "
            "configuration argument."
            " Anions must have a negative charge.",
        ):
            m.an = Anion(charge=+1, _electrolyte=True)

    @pytest.mark.unit
    def test_no_charge(self, m):
        with pytest.raises(
            ConfigurationError, match="an was not provided with a value for charge."
        ):
            m.an = Anion(_electrolyte=True)

    @pytest.mark.unit
    def test_is_phase_valid_no_assignment(self, m):
        with pytest.raises(AttributeError):
            m.comp._is_phase_valid("foo")

    @pytest.mark.unit
    def test_is_phase_valid_liquid(self, m):
        m.Liq = LiquidPhase()
        m.Sol = SolidPhase()
        m.Vap = VaporPhase()
        m.Aqu = AqueousPhase()
        m.Phase = Phase()

        assert not m.comp._is_phase_valid(m.Liq)
        assert not m.comp._is_phase_valid(m.Sol)
        assert not m.comp._is_phase_valid(m.Vap)
        assert m.comp._is_phase_valid(m.Aqu)
        assert not m.comp._is_phase_valid(m.Phase)


class TestCation:
    @pytest.mark.unit
    def test_not_electrolyte(self):
        m = ConcreteModel()

        m.meta_object = PropertyClassMetadata()

        def get_metadata(self):
            return m.meta_object

        m.get_metadata = types.MethodType(get_metadata, m)

        with pytest.raises(
            PropertyPackageError,
            match="comp Ion Component types should only be used with Aqueous Phases",
        ):
            m.comp = Cation(_electrolyte=False)

    @pytest.fixture(scope="class")
    def m(self):
        m = ConcreteModel()

        m.meta_object = PropertyClassMetadata()
        m.meta_object._default_units.set_units(
            temperature=pyunits.K,
            mass=pyunits.kg,
            length=pyunits.m,
            time=pyunits.s,
            amount=pyunits.mol,
        )

        def get_metadata(self):
            return m.meta_object

        m.get_metadata = types.MethodType(get_metadata, m)

        m.cation_set = Set()

        m.comp = Cation(charge=+1, _electrolyte=True)

        return m

    @pytest.mark.unit
    def test_config(self, m):
        assert "valid_phase_types" not in m.comp.config
        assert m.comp.config.charge == +1
        assert not m.comp.config._component_list_exists
        assert not m.comp.config.has_vapor_pressure
        with pytest.raises(ValueError):
            m.comp.config.has_vapor_pressure = True

    @pytest.mark.unit
    def test_populate_component_list(self, m):
        assert isinstance(m.cation_set, Set)

        for j in m.cation_set:
            assert j in ["comp"]

    @pytest.mark.unit
    def test_is_solute(self, m):
        assert m.comp.is_solute()

    @pytest.mark.unit
    def test_is_solvent(self, m):
        assert not m.comp.is_solvent()

    @pytest.mark.unit
    def test_invalid_charge(self, m):
        with pytest.raises(
            ConfigurationError,
            match="cat received invalid value for charge "
            "configuration argument."
            " Cations must have a positive charge.",
        ):
            m.cat = Cation(charge=-1, _electrolyte=True)

    @pytest.mark.unit
    def test_no_charge(self, m):
        with pytest.raises(
            ConfigurationError, match="cat was not provided with a value for charge."
        ):
            m.cat = Cation(_electrolyte=True)

    @pytest.mark.unit
    def test_is_phase_valid_no_assignment(self, m):
        with pytest.raises(AttributeError):
            m.comp._is_phase_valid("foo")

    @pytest.mark.unit
    def test_is_phase_valid_liquid(self, m):
        m.Liq = LiquidPhase()
        m.Sol = SolidPhase()
        m.Vap = VaporPhase()
        m.Aqu = AqueousPhase()
        m.Phase = Phase()

        assert not m.comp._is_phase_valid(m.Liq)
        assert not m.comp._is_phase_valid(m.Sol)
        assert not m.comp._is_phase_valid(m.Vap)
        assert m.comp._is_phase_valid(m.Aqu)
        assert not m.comp._is_phase_valid(m.Phase)


class TestApparent:
    @pytest.fixture(scope="class")
    def m(self):
        m = ConcreteModel()

        m.meta_object = PropertyClassMetadata()
        m.meta_object._default_units.set_units(
            temperature=pyunits.K,
            mass=pyunits.kg,
            length=pyunits.m,
            time=pyunits.s,
            amount=pyunits.mol,
        )

        def get_metadata(self):
            return m.meta_object

        m.get_metadata = types.MethodType(get_metadata, m)

        m._apparent_set = Set()

        m.comp = Apparent(dissociation_species={"comp": 1}, _electrolyte=True)

        return m

    @pytest.mark.unit
    def test_not_electrolyte(self):
        m = ConcreteModel()

        m.meta_object = PropertyClassMetadata()
        m.meta_object._default_units.set_units(
            temperature=pyunits.K,
            mass=pyunits.kg,
            length=pyunits.m,
            time=pyunits.s,
            amount=pyunits.mol,
        )

        def get_metadata(self):
            return m.meta_object

        m.get_metadata = types.MethodType(get_metadata, m)

        m.comp = Apparent(dissociation_species={"comp": 1}, _electrolyte=False)

        for j in m.component_list:
            assert j in ["comp"]
        for j in m.solute_set:
            assert j in ["comp"]

        assert m.comp.config.dissociation_species == {"comp": 1}

    @pytest.mark.unit
    def test_no_dissociation_species(self):
        m = ConcreteModel()

        m.meta_object = PropertyClassMetadata()
        m.meta_object._default_units.set_units(
            temperature=pyunits.K,
            mass=pyunits.kg,
            length=pyunits.m,
            time=pyunits.s,
            amount=pyunits.mol,
        )

        def get_metadata(self):
            return m.meta_object

        m.get_metadata = types.MethodType(get_metadata, m)

        with pytest.raises(
            ConfigurationError,
            match="dissoication_species argument was not set. "
            "Apparent components require the dissociation species to be "
            "defined.",
        ):
            m.comp = Apparent(_electrolyte=False)

    @pytest.mark.unit
    def test_config(self, m):
        assert m.comp.config.valid_phase_types is None
        assert not m.comp.config._component_list_exists

    @pytest.mark.unit
    def test_populate_component_list(self, m):
        assert isinstance(m._apparent_set, Set)

        for j in m._apparent_set:
            assert j in ["comp"]

    @pytest.mark.unit
    def test_is_solute(self, m):
        assert m.comp.is_solute()

    @pytest.mark.unit
    def test_is_solvent(self, m):
        assert not m.comp.is_solvent()

    @pytest.mark.unit
    def test_is_phase_valid_no_assignment(self, m):
        with pytest.raises(TypeError):
            m.comp._is_phase_valid("foo")

    @pytest.mark.unit
    def test_is_phase_valid_liquid(self, m):
        m.comp3 = Apparent(
            valid_phase_types=PhaseType.liquidPhase,
            dissociation_species={"comp": 1},
            _electrolyte=True,
        )

        m.Liq = LiquidPhase()
        m.Sol = SolidPhase()
        m.Vap = VaporPhase()
        m.Aqu = AqueousPhase()
        m.Phase = Phase()

        assert m.comp3._is_phase_valid(m.Liq)
        assert not m.comp3._is_phase_valid(m.Sol)
        assert not m.comp3._is_phase_valid(m.Vap)
        assert not m.comp3._is_phase_valid(m.Aqu)
        assert not m.comp3._is_phase_valid(m.Phase)

    @pytest.mark.unit
    def test_is_phase_valid_vapor(self, m):
        m.comp4 = Apparent(
            valid_phase_types=PhaseType.vaporPhase,
            dissociation_species={"comp": 1},
            _electrolyte=True,
        )

        assert not m.comp4._is_phase_valid(m.Liq)
        assert not m.comp4._is_phase_valid(m.Sol)
        assert m.comp4._is_phase_valid(m.Vap)
        assert not m.comp4._is_phase_valid(m.Aqu)
        assert not m.comp4._is_phase_valid(m.Phase)

    @pytest.mark.unit
    def test_is_phase_valid_solid(self, m):
        m.comp5 = Apparent(
            valid_phase_types=PhaseType.solidPhase,
            dissociation_species={"comp": 1},
            _electrolyte=True,
        )

        assert not m.comp5._is_phase_valid(m.Liq)
        assert m.comp5._is_phase_valid(m.Sol)
        assert not m.comp5._is_phase_valid(m.Vap)
        assert not m.comp5._is_phase_valid(m.Aqu)
        assert not m.comp5._is_phase_valid(m.Phase)

    @pytest.mark.unit
    def test_is_phase_valid_aqueous(self, m):
        m.comp6 = Apparent(
            valid_phase_types=PhaseType.aqueousPhase,
            dissociation_species={"comp": 1},
            _electrolyte=True,
        )

        assert not m.comp6._is_phase_valid(m.Liq)
        assert not m.comp6._is_phase_valid(m.Sol)
        assert not m.comp6._is_phase_valid(m.Vap)
        assert m.comp6._is_phase_valid(m.Aqu)
        assert not m.comp6._is_phase_valid(m.Phase)

    @pytest.mark.unit
    def test_is_phase_valid_LV(self, m):
        m.comp7 = Apparent(
            valid_phase_types=[PhaseType.liquidPhase, PhaseType.vaporPhase],
            dissociation_species={"comp": 1},
            _electrolyte=True,
        )

        assert m.comp7._is_phase_valid(m.Liq)
        assert not m.comp7._is_phase_valid(m.Sol)
        assert m.comp7._is_phase_valid(m.Vap)
        assert not m.comp7._is_phase_valid(m.Aqu)
        assert not m.comp7._is_phase_valid(m.Phase)

    @pytest.mark.unit
    def test_is_aqueous_phase_valid(self, m):
        assert m.comp._is_aqueous_phase_valid()



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################

from idaes.models_extra.power_generation.unit_models.soc_submodels.channel import (
    SocChannel,
)
from idaes.models_extra.power_generation.unit_models.soc_submodels.porous_conductive_slab import (
    PorousConductiveSlab,
)
from idaes.models_extra.power_generation.unit_models.soc_submodels.conductive_slab import (
    SocConductiveSlab,
)
from idaes.models_extra.power_generation.unit_models.soc_submodels.triple_phase_boundary import (
    SocTriplePhaseBoundary,
)
from idaes.models_extra.power_generation.unit_models.soc_submodels.contact_resistor import (
    SocContactResistor,
)
from idaes.models_extra.power_generation.unit_models.soc_submodels.solid_oxide_cell import (
    SolidOxideCell,
)
from idaes.models_extra.power_generation.unit_models.soc_submodels.solid_oxide_module_simple import (
    SolidOxideModuleSimple,
)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
# TODO: Missing doc strings
# pylint: disable=missing-module-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=missing-function-docstring

import itertools

import numpy as np
import pandas as pd
import idaes.logger as idaeslog

_log = idaeslog.getLogger(__name__)

__author__ = "Oluwamayowa Amusat"


class FeatureScaling:
    """

    A class for scaling and unscaling input and output data. The class contains three main functions
    """

    def __init__(self):
        pass

    @staticmethod
    def data_scaling_minmax(data):
        """

        This function performs column-wise minimax scaling on the input dataset.

            Args:
                data (NumPy Array or Pandas Dataframe): The input data set to be scaled. Must be a numpy array or dataframe.

            Returns:
                scaled_data(NumPy Array): A 2-D numpy array containing the scaled data. All array values will be between [0, 1].
                data_minimum(NumPy Array): A 2-D row vector containing the column-wise minimums of the input data
                data_maximum(NumPy Array): A 2-D row vector containing the column-wise maximums of the input data

            Raises:
                TypeError: Raised when the input data is not a numpy array or dataframe
        """
        # Confirm that data type is an array or DataFrame
        if isinstance(data, np.ndarray):
            input_data = data
        elif isinstance(data, pd.DataFrame):
            input_data = data.values
        else:
            raise TypeError(
                "original_data_input: Pandas dataframe or numpy array required."
            )

        if input_data.ndim == 1:
            input_data = input_data.reshape(len(input_data), 1)
        data_minimum = np.min(input_data, axis=0)
        data_maximum = np.max(input_data, axis=0)
        scale = data_maximum - data_minimum
        scale[scale == 0.0] = 1.0
        scaled_data = (input_data - data_minimum) / scale
        # scaled_data = (input_data - data_minimum) / (data_maximum - data_minimum)
        data_minimum = data_minimum.reshape(1, data_minimum.shape[0])
        data_maximum = data_maximum.reshape(1, data_maximum.shape[0])
        return scaled_data, data_minimum, data_maximum

    @staticmethod
    def data_unscaling_minmax(x_scaled, x_min, x_max):
        """

        This function performs column-wise un-scaling on the a minmax-scaled input dataset.

            Args:
                x_scaled(NumPy Array): The input data set to be un-scaled. Data values should be between 0 and 1.
                x_min(NumPy Array): 1-D or 2-D (n-by-1) vector containing the actual minimum value for each column. Must contain same number of elements as the number of columns in x_scaled.
                x_max(NumPy Array): 1-D or 2-D (n-by-1) vector containing the actual maximum value for each column. Must contain same number of elements as the number of columns in x_scaled.

            Returns:
                unscaled_data(NumPy Array): A 2-D numpy array containing the scaled data, unscaled_data = x_min + x_scaled * (x_max - x_min)

            Raises:
                IndexError: Function raises index error when the dimensions of the arrays are inconsistent.
        """
        # Check if it can be evaluated. Will return index error if dimensions are wrong
        if x_scaled.ndim == 1:  # Check if 1D, and convert to 2D if required.
            x_scaled = x_scaled.reshape(len(x_scaled), 1)
        if (x_scaled.shape[1] != x_min.size) or (x_scaled.shape[1] != x_max.size):
            raise IndexError("Dimensionality problems with data for un-scaling.")
        unscaled_data = x_min + x_scaled * (x_max - x_min)
        return unscaled_data

    # @staticmethod
    # def data_scaling_standardization(data):
    #     # Confirm that data type is an array or DataFrame
    #     if isinstance(data, np.ndarray):
    #         input_data = data
    #     elif isinstance(data, pd.DataFrame):
    #         input_data = data.values
    #     else:
    #         raise TypeError('original_data_input: Pandas dataframe or numpy array required.')
    #
    #     if input_data.ndim == 1:
    #         input_data = input_data.reshape(len(input_data), 1)
    #
    #     data_mean = np.mean(input_data, axis=0)
    #     data_stdev = np.std(input_data, axis=0)
    #     scaled_data = (input_data - data_mean) / data_stdev
    #     data_mean = data_mean.reshape(1, data_mean.shape[0])
    #     data_stdev = data_stdev.reshape(1, data_stdev.shape[0])
    #     return scaled_data, data_mean, data_stdev


class SamplingMethods:
    def nearest_neighbour(self, full_data, a):
        """
        Function determines the closest point to a in data_input (user provided data).
        This is done by determining the input data with the smallest L2 distance from a.

        The function:
        1. Calculates the L2 distance between all the input data points and a,
        2. Sorts the input data based on the calculated L2-distances, and
        3. Selects the sample point in the first row (after sorting) as the closest sample point.

        Args:
            self: contains, among other things, the input data.
            full_data: refers to the input dataset supplied by the user.
            a: a single row vector containing the sample point we want to find the closest sample to.

        Returns:
            closest_point: a row vector containing the closest point to a in self.x_data
        """
        no_y_vars = self.x_data.shape[1] - full_data.shape[1]
        dist = full_data[:, :no_y_vars] - a
        l2_norm = np.sqrt(np.sum((dist**2), axis=1))
        l2_norm = l2_norm.reshape(l2_norm.shape[0], 1)
        distances = np.append(full_data, l2_norm, 1)
        sorted_distances = distances[distances[:, -1].argsort()]
        closest_point = sorted_distances[0, :-1]
        return closest_point

    def points_selection(self, full_data, generated_sample_points):
        """
        Uses L2-distance evaluation (implemented in nearest_neighbour) to find closest available points in original data to those generated by the sampling technique.
        Calls the nearest_neighbour function for each row in the input data.

        Args:
            full_data: refers to the input dataset supplied by the user.
            generated_sample_points(NumPy Array): The vector of points (number_of_sample rows) for which the closest points in the original data are to be found. Each row represents a sample point.

        Returns:
            equivalent_points: Array containing the points (in rows) most similar to those in generated_sample_points
        """

        equivalent_points = np.zeros(
            (generated_sample_points.shape[0], len(self.data_headers))
        )
        for i in range(0, generated_sample_points.shape[0]):
            closest_point = self.nearest_neighbour(
                full_data, generated_sample_points[i, :]
            )
            equivalent_points[i, :] = closest_point
        return equivalent_points

    def sample_point_selection(self, full_data, sample_points, sampling_type):
        if sampling_type == "selection":
            sd = FeatureScaling()
            scaled_data, data_min, data_max = sd.data_scaling_minmax(full_data)
            points_closest_scaled = self.points_selection(scaled_data, sample_points)
            points_closest_unscaled = sd.data_unscaling_minmax(
                points_closest_scaled, data_min, data_max
            )

            unique_sample_points = np.unique(points_closest_unscaled, axis=0)
            if unique_sample_points.shape[0] < points_closest_unscaled.shape[0]:
                _log.warning(
                    "The returned number of samples is less than the requested number due to repetitions during nearest neighbour selection."
                )
            print(
                "\nNumber of unique samples returned by sampling algorithm:",
                unique_sample_points.shape[0],
            )

        elif sampling_type == "creation":
            sd = FeatureScaling()
            unique_sample_points = sd.data_unscaling_minmax(
                sample_points, full_data[0, :], full_data[1, :]
            )

        return unique_sample_points

    def prime_number_generator(self, n):
        """
        Function generates a list of the first n prime numbers

            Args:
                n(int): Number of prime numbers required

            Returns:
                prime_list(list): A list of the first n prime numbers

        Example: Generate first three prime numbers
            >>  prime_number_generator(3)
            >> [2, 3, 5]

        """
        # Alternative way of generating primes using list generators
        # prime_list = []
        # current_no = 2
        # while len(prime_list) < n:
        #     matching_objs = next((o for o in range(2, current_no) if current_no % o == 0), 0)
        #     if matching_objs==0:
        #         prime_list.append(current_no)
        #     current_no += 1

        prime_list = []
        current_no = 2
        while len(prime_list) < n:
            for i in range(2, current_no):
                if (current_no % i) == 0:
                    break
            else:
                prime_list.append(current_no)
            current_no += 1
        return prime_list

    def base_conversion(self, a, b):
        """
        Function converts integer a from base 10 to base b

            Args:
                a(int): Number to be converted, base 10
                b(int): Base required

            Returns:
                string_representation(list): List containing strings of individual digits of "a" in the new base "b"

        Examples: Convert (i) 5 to base 2 and (ii) 57 to base 47
            >>  base_conversion(5, 2)
            >> ['1', '0', '1']

            >>  base_conversion(57, 47)
            >> ['1', '10']

        """

        string_representation = []
        if a < b:
            string_representation.append(str(a))
        else:
            while a > 0:
                a, c = (a // b, a % b)
                string_representation.append(str(c))
            string_representation = string_representation[::-1]
        return string_representation

    def prime_base_to_decimal(self, num, base):
        """
        ===============================================================================================================
        Function converts a fractional number "num" in base "base" to base 10. Reverses the process in base_conversion
        Note: The first string element is ignored, since this would be zero for a fractional number.

            Args:
                num(list): Number in base b to be converted. The number must be represented as a list containing individual digits of the base, with the first entry as zero.
                b(int): Original base

            Returns:
                decimal_equivalent(float): Fractional number in base 10

        Examples:
        Convert 0.01 (base 2) to base 10
            >>  prime_base_to_decimal(['0', '0', '1'], 2)  # Represents 0.01 in base 2
            >> 0.25

        Convert 0.01 (base 20) to base 10
            >>  prime_base_to_decimal(['0', '0', '1'], 20)  # Represents 0.01 in base 20
            >> 0.0025
        ================================================================================================================

        """
        binary = num
        decimal_equivalent = 0
        # Convert fractional part decimal equivalent
        for i in range(1, len(binary)):
            decimal_equivalent += int(binary[i]) / (base**i)
        return decimal_equivalent

    def data_sequencing(self, no_samples, prime_base):
        """
        ===============================================================================================================
        Function which generates the first no_samples elements of the Halton or Hammersley sequence based on the prime number prime_base
        The steps for generating the first no_samples of the sequence are as follows:
        1. Create a list of numbers between 0 and no_samples --- nums = [0, 1, 2, ..., no_samples]
        2. Convert each element in nums into its base form based on the prime number prime_base, reverse the base digits of each number in num
        3. Add a decimal point in front of the reversed number
        4. Convert the reversed numbers back to base 10
            Args:
                no_samples(int): Number of Halton/Hammersley sequence elements required
                prime_base(int): Current prime number to be used as base

            Returns:
                sequence_decimal(NumPy Array): 1-D array containing the first no_samples elements of the sequence based on prime_base

        Examples:
        First three elements of the Halton sequence based on base 2
            >>  data_sequencing(self, 3, 2)
            >> [0, 0.5, 0.75]
        ================================================================================================================

        """
        pure_numbers = np.arange(0, no_samples)
        bitwise_rep = []
        reversed_bitwise_rep = []
        sequence_bitwise = []
        sequence_decimal = np.zeros((no_samples, 1))
        for i in range(0, no_samples):
            base_rep = self.base_conversion(pure_numbers[i], prime_base)
            bitwise_rep.append(base_rep)
            reversed_bitwise_rep.append(base_rep[::-1])
            sequence_bitwise.append(["0."] + reversed_bitwise_rep[i])
            sequence_decimal[i, 0] = self.prime_base_to_decimal(
                sequence_bitwise[i], prime_base
            )
        sequence_decimal = sequence_decimal.reshape(
            sequence_decimal.shape[0],
        )
        return sequence_decimal

    def selection_columns_preprocessing(self, data_input, xlabels, ylabels):
        """
        Pre-processing data for multiple output selection case.

        Args:
            data_input:     data supplied by user (dataframe or numpy array)
            xlabels:        list of input variables
            ylabels:        list of output variables
        """
        self.df_flag = True

        if isinstance(data_input, pd.DataFrame):
            if xlabels is None:
                xlabels = []
            if ylabels is None:
                ylabels = []
            set_of_labels = xlabels + ylabels

            if not set_of_labels:
                # Empty list assumes one output in final column - all other variables treated as inputs
                print(
                    "No column information provided. All except last column will be considered as x variables."
                )
                self.data = data_input.values
                self.data_headers = data_input.columns.values.tolist()
                self.x_data = self.data[:, :-1]
                self.data_headers_xvars = self.data_headers[:-1]  # Last column is y
            elif not set(set_of_labels).issubset(data_input.columns):
                raise IndexError(
                    'Invalid column names in "xlabels" or "ylabels": ',
                    [k for k in set_of_labels if k not in data_input.columns],
                )
            elif not ylabels:
                # When only xlabels are supplied, assume all non-inputs are outputs
                self.x_data = data_input.filter(xlabels).values
                self.data_headers_xvars = xlabels
                non_x_vars = [k for k in data_input.columns if k not in xlabels]
                y_data = data_input.filter(non_x_vars).values
                self.data_headers = self.data_headers_xvars + non_x_vars
                self.data = np.concatenate((self.x_data, y_data), axis=1)
            elif not xlabels:
                # When only ylabels are supplied, assume all non-outputs are inputs
                xvars = [k for k in data_input.columns if k not in ylabels]
                self.x_data = data_input.filter(xvars).values
                self.data_headers_xvars = xvars
                y_data = data_input.filter(ylabels).values
                self.data_headers = self.data_headers_xvars + ylabels
                self.data = np.concatenate((self.x_data, y_data), axis=1)
            else:
                # Only provided columns are considered - any column not in xlabels and ylabels is dropped completely.
                dropped_cols = [k for k in data_input.columns if k not in set_of_labels]
                if len(dropped_cols) > 0:
                    warn_str = "The following columns were dropped: " + str(
                        dropped_cols
                    )
                    _log.warning(warn_str)
                self.x_data = data_input.filter(xlabels).values
                self.data_headers = set_of_labels
                self.data_headers_xvars = xlabels
                self.data = data_input.filter(set_of_labels).values

        elif isinstance(data_input, np.ndarray):
            self.df_flag = False

            if xlabels is None:
                xlabels = []
            if ylabels is None:
                ylabels = []
            set_of_labels = xlabels + ylabels

            if not set_of_labels:
                # Empty list assumes one output in final column
                print(
                    "No column information provided. All except last column will be considered as x variables."
                )
                self.data = data_input
                self.data_headers = [i for i in range(0, data_input.shape[1])]
                self.x_data = self.data[:, :-1]
                self.data_headers_xvars = self.data_headers[:-1]  # Last column is y
            elif not set(set_of_labels).issubset(range(0, data_input.shape[1])):
                raise IndexError(
                    'Invalid column names in "xlabels" or "ylabels": ',
                    [
                        k
                        for k in set_of_labels
                        if k not in range(0, data_input.shape[1])
                    ],
                )
            elif not ylabels:
                # Assumes all non-inputs are outputs
                self.x_data = data_input[:, xlabels]
                self.data_headers_xvars = xlabels
                non_x_vars = [
                    k for k in range(0, data_input.shape[1]) if k not in xlabels
                ]
                y_data = data_input[:, non_x_vars]
                self.data_headers = self.data_headers_xvars + non_x_vars
                self.data = np.concatenate((self.x_data, y_data), axis=1)
            elif not xlabels:
                # Assumes all non-outputs are inputs
                xvars = [k for k in range(0, data_input.shape[1]) if k not in ylabels]
                self.x_data = data_input[:, xvars]
                self.data_headers_xvars = xvars
                y_data = data_input[:, ylabels]
                self.data_headers = self.data_headers_xvars + ylabels
                self.data = np.concatenate((self.x_data, y_data), axis=1)
            else:
                # Only provided columns are considered
                dropped_cols = [
                    k for k in range(0, data_input.shape[1]) if k not in set_of_labels
                ]
                if len(dropped_cols) > 0:
                    warn_str = "The following columns were dropped: " + str(
                        dropped_cols
                    )
                    _log.warning(warn_str)
                self.x_data = data_input[:, xlabels]
                self.data_headers = set_of_labels
                self.data_headers_xvars = xlabels
                self.data = data_input[:, set_of_labels]


class LatinHypercubeSampling(SamplingMethods):
    """
    A class that performs Latin Hypercube Sampling. The function returns LHS samples which have been selected randomly after sample space stratification.

    It should be noted that no minimax criterion has been used in this implementation, so the LHS samples selected will not have space-filling properties.

    To use: call class with inputs, and then run ``sample_points`` method.

    **Example:**

    .. code-block:: python

        # To select 10 LHS samples from "data"
        >>> b = rbf.LatinHypercubeSampling(data, 10, sampling_type="selection")
        >>> samples = b.sample_points()

    """

    def __init__(
        self,
        data_input,
        number_of_samples=None,
        sampling_type=None,
        xlabels=None,
        ylabels=None,
        rand_seed=None,
    ):
        """
        Initialization of **LatinHypercubeSampling** class. Two inputs are required.

        Args:
            data_input (NumPy Array, Pandas Dataframe or list) :  The input data set or range to be sampled.

                - When the aim is to select a set of samples from an existing dataset, the dataset must be a NumPy Array or a Pandas Dataframe and **sampling_type** option must be set to "selection". A single output variable (y) is assumed to be supplied in the last column if **xlabels** and **ylabels** are not supplied.
                - When the aim is to generate a set of samples from a data range, the dataset must be a list containing two lists of equal lengths which contain the variable bounds and **sampling_type** option must be set to "creation". It is assumed that the range contains no output variable information in this case.

            number_of_samples (int): The number of samples to be generated. Should be a positive integer less than or equal to the number of entries (rows) in **data_input**.
            sampling_type (str) : Option which determines whether the algorithm selects samples from an existing dataset ("selection") or attempts to generate sample from a supplied range ("creation"). Default is "creation".

        Keyword Args:
            xlabels (list): List of column names (if **data_input** is a dataframe) or column numbers (if **data_input** is an array) for the independent/input  variables.  Only used in "selection" mode. Default is None.
            ylabels (list): List of column names (if **data_input** is a dataframe) or column numbers (if **data_input** is an array) for the dependent/output variables. Only used in "selection" mode. Default is None.
            rand_seed (int): Option that allows users to fix the numpy random seed generator for reproducibility (if required).

        Returns:
            **self** function containing the input information

        Raises:
            ValueError: The input data (**data_input**) is the wrong type/dimension, or **number_of_samples** is invalid (too large, zero, or negative)

            TypeError: When **number_of_samples** is not the right type, or **sampling_type** entry is not a string.

            IndexError: When invalid column names are supplied in **xlabels** or **ylabels**

        """

        if sampling_type is None:
            sampling_type = "creation"
            self.sampling_type = sampling_type
            print("Creation-type sampling will be used.")
        elif not isinstance(sampling_type, str):
            raise TypeError("Invalid sampling type entry. Must be of type <str>.")
        elif (sampling_type.lower() == "creation") or (
            sampling_type.lower() == "selection"
        ):
            sampling_type = sampling_type.lower()
            self.sampling_type = sampling_type
        else:
            raise ValueError(
                'Invalid sampling type requirement entered. Enter "creation" for sampling from a range or "selection" for selecting samples from a dataset.'
            )
        print("Sampling type: ", self.sampling_type, "\n")

        if self.sampling_type == "selection":
            if isinstance(data_input, (pd.DataFrame, np.ndarray)):
                self.selection_columns_preprocessing(data_input, xlabels, ylabels)
            else:
                raise ValueError(
                    'Pandas dataframe or numpy array required for sampling_type "selection."'
                )

            # Catch potential errors in number_of_samples
            if number_of_samples is None:
                print(
                    "\nNo entry for number of samples to be generated. The default value of 5 will be used."
                )
                number_of_samples = 5
            elif number_of_samples > self.data.shape[0]:
                raise ValueError(
                    "LHS sample size cannot be greater than number of samples in the input data set"
                )
            elif not isinstance(number_of_samples, int):
                raise TypeError("number_of_samples must be an integer.")
            elif number_of_samples <= 0:
                raise ValueError("number_of_samples must a positive, non-zero integer.")
            self.number_of_samples = number_of_samples

        elif self.sampling_type == "creation":
            if not isinstance(data_input, list):
                raise ValueError(
                    'List entry of two elements expected for sampling_type "creation."'
                )
            elif (
                len(data_input) != 2
                or not isinstance(data_input[0], list)
                or not isinstance(data_input[1], list)
                or len(data_input[0]) != len(data_input[1])
            ):
                raise ValueError("data_input must contain two lists of equal lengths.")
            elif data_input[0] == data_input[1]:
                raise ValueError("Invalid entry: both lists are equal.")
            elif any(x == y for x, y in zip(data_input[0], data_input[1])):
                raise ValueError(
                    "Invalid entry: at least one variable contains the same value for the lower and upper bounds."
                )
            else:
                bounds_array = np.zeros(
                    (
                        2,
                        len(data_input[0]),
                    )
                )
                bounds_array[0, :] = np.array(data_input[0])
                bounds_array[1, :] = np.array(data_input[1])
                data_headers = []
            self.data = bounds_array
            self.data_headers = data_headers

            # Catch potential errors in number_of_samples
            if number_of_samples is None:
                print(
                    "\nNo entry for number of samples to be generated. The default value of 5 will be used."
                )
                number_of_samples = 5
            elif not isinstance(number_of_samples, int):
                raise TypeError("number_of_samples must be an integer.")
            elif number_of_samples <= 0:
                raise ValueError("number_of_samples must a positive, non-zero integer.")
            self.number_of_samples = number_of_samples
            self.x_data = bounds_array  # Only x data will be present in this case

        if rand_seed is not None:
            try:
                self.seed_value = int(rand_seed)
                np.random.seed(self.seed_value)
            except ValueError:
                raise ValueError("Random seed must be an integer.")

    def variable_sample_creation(self, variable_min, variable_max):
        """

        Function that generates the required number of sample points for a given variable within a specified range using stratification.
        The function divides the variable sample space into self.number_of_samples equal strata and generates a single random sample from each strata based on its lower and upper bound.

        Args:
            self
            variable_min(float): The lower bound of the sample space region. Should be a single number.
            variable_max(float): The upper bound of the sample space region. Should be a single number.

        Returns:
            var_samples(NumPy Array): A numpy array of size (self.number_of_samples x 1) containing the randomly generated points from each strata
        """

        strata_size = 1 / self.number_of_samples
        var_samples = np.zeros((self.number_of_samples, 1))
        for i in range(self.number_of_samples):
            strata_lb = i * strata_size
            sample_point = strata_lb + (np.random.rand() * strata_size)
            var_samples[i, 0] = (
                sample_point * (variable_max - variable_min)
            ) + variable_min
        return var_samples

    def lhs_points_generation(self):
        """
        Generate points within each strata for each variable based on stratification. When invoked, it:
        1. Determines the mimumum and maximum value for each feature (column),
        2. Calls the variable_sample_creation function on each feature, passing in its mimmum and maximum
        3. Returns an array containing the points selected in each strata of each column

        Returns:
            sample_points_vector(NumPy Array): Array containing the columns of the random samples generated in each strata.
        """

        ns, nf = np.shape(self.x_data)  # pylint: disable=unused-variable
        sample_points_vector = np.zeros(
            (self.number_of_samples, nf)
        )  # Array containing points in each interval for each variable
        for i in range(nf):
            variable_min = 0  # np.min(self.x_data[:, i])
            variable_max = 1  # np.max(self.x_data[:, i])
            var_samples = self.variable_sample_creation(
                variable_min, variable_max
            )  # Data generation step
            sample_points_vector[:, i] = var_samples[:, 0]
        return sample_points_vector

    @staticmethod
    def random_shuffling(vector_of_points):
        """
        This function carries out random shuffling of column data to generate samples.
        Data in each of the columns  in the input array is shuffled separately, meaning that the rows of the resultant array will contain random samples from the sample space.

        Args:
            vector_of_points(NumPy Array): Array containing ordered points generated from stratification. Should usually be the output of the lhs_points_generation function. Each column self.number_of_samples elements.

        Returns:
            vector_of_points(NumPy Array): 2-D array containing the shuffled data. Should contain number_of_sample rows, with each row representing a potential random sample from within the sample space.

        """

        _, nf = np.shape(vector_of_points)
        for i in range(0, nf):
            z_col = vector_of_points[:, i]
            np.random.shuffle(z_col)
            vector_of_points[:, i] = z_col
        return vector_of_points

    def sample_points(self):
        """
        ``sample_points`` generates or selects Latin Hypercube samples from an input dataset or data range. When called, it:

            1. generates samples points from stratified regions by calling the ``lhs_points_generation``,
            2. generates potential sample points by random shuffling, and
            3. when a dataset is provided, selects the closest available samples to the theoretical sample points from within the input data.

        Returns:
            NumPy Array or Pandas Dataframe:     A numpy array or Pandas dataframe containing **number_of_samples** points selected or generated by LHS.

        """

        vector_of_points = (
            self.lhs_points_generation()
        )  # Assumes [X, Y] data is supplied.
        generated_sample_points = self.random_shuffling(vector_of_points)
        unique_sample_points = self.sample_point_selection(
            self.data, generated_sample_points, self.sampling_type
        )

        if len(self.data_headers) > 0 and self.df_flag:
            unique_sample_points = pd.DataFrame(
                unique_sample_points, columns=self.data_headers
            )
        return unique_sample_points


class UniformSampling(SamplingMethods):
    """
    A class that performs Uniform Sampling. Depending on the settings, the algorithm either returns samples from an input dataset  which have been selected using Euclidean distance minimization after the uniform samples have been generated,
    or returns samples from a supplied data range.

    Full-factorial samples are based on dividing the space of each variable randomly and then generating all possible variable combinations.

    - The number of points to be sampled per variable needs to be specified in a list.

    To use: call class with inputs, and then ``sample_points`` function

    **Example:**

    .. code-block:: python

        # To select 50 samples on a (10 x 5) grid in a 2D space:
        >>> b = rbf.UniformSampling(data, [10, 5], sampling_type="selection")
        >>> samples = b.sample_points()

    """

    def __init__(
        self,
        data_input,
        list_of_samples_per_variable,
        sampling_type=None,
        xlabels=None,
        ylabels=None,
        edges=None,
    ):
        """
        Initialization of UniformSampling class. Three inputs are required.

        Args:
            data_input (NumPy Array, Pandas Dataframe or list) :  The input data set or range to be sampled.

                - When the aim is to select a set of samples from an existing dataset, the dataset must be a NumPy Array or a Pandas Dataframe and **sampling_type** option must be set to "selection". A single output variable (y) is assumed to be supplied in the last column if **xlabels** and **ylabels** are not supplied.
                - When the aim is to generate a set of samples from a data range, the dataset must be a list containing two lists of equal lengths which contain the variable bounds and **sampling_type** option must be set to "creation". It is assumed that the range contains no output variable information in this case.

            list_of_samples_per_variable (list): The list containing the number of subdivisions for each variable. Each dimension (variable) must be represented by a positive integer variable greater than 1.
            sampling_type (str) : Option which determines whether the algorithm selects samples from an existing dataset ("selection") or attempts to generate sample from a supplied range ("creation"). Default is "creation".

        Keyword Args:
            xlabels (list): List of column names (if **data_input** is a dataframe) or column numbers (if **data_input** is an array) for the independent/input  variables.  Only used in "selection" mode. Default is None.
            ylabels (list): List of column names (if **data_input** is a dataframe) or column numbers (if **data_input** is an array) for the dependent/output variables. Only used in "selection" mode. Default is None.
            edges (bool): Boolean variable representing how the points should be selected. A value of True (default) indicates the points should be equally spaced edge to edge, otherwise they will be in the centres of the bins filling the unit cube

        Returns:
            **self** function containing the input information

        Raises:
            ValueError: The **data_input** is the wrong type, or **list_of_samples_per_variable** is of the wrong length, or **list_of_samples_per_variable** is invalid.

            TypeError: When **list_of_samples_per_variable** is not a list,  or **list_of_samples_per_variable** contains elements other than integers, **sampling_type** is not a string, or **edges** entry is not Boolean

            IndexError: When invalid column names are supplied in **xlabels** or **ylabels**

        """
        if sampling_type is None:
            sampling_type = "creation"
            self.sampling_type = sampling_type
            print("Creation-type sampling will be used.")
        elif not isinstance(sampling_type, str):
            raise TypeError("Invalid sampling type entry. Must be of type <str>.")
        elif (sampling_type.lower() == "creation") or (
            sampling_type.lower() == "selection"
        ):
            sampling_type = sampling_type.lower()
            self.sampling_type = sampling_type
        else:
            raise ValueError(
                'Invalid sampling type requirement entered. Enter "creation" for sampling from a range or "selection" for selecting samples from a dataset.'
            )
        print("Sampling type: ", self.sampling_type, "\n")

        if self.sampling_type == "selection":
            if isinstance(data_input, (pd.DataFrame, np.ndarray)):
                self.selection_columns_preprocessing(data_input, xlabels, ylabels)
            else:
                raise ValueError(
                    'Pandas dataframe or numpy array required for sampling_type "selection."'
                )

        elif self.sampling_type == "creation":
            if not isinstance(data_input, list):
                raise ValueError(
                    'List entry of two elements expected for sampling_type "creation."'
                )
            elif (
                len(data_input) != 2
                or not isinstance(data_input[0], list)
                or not isinstance(data_input[1], list)
                or len(data_input[0]) != len(data_input[1])
            ):
                raise ValueError("data_input must contain two lists of equal lengths.")
            elif data_input[0] == data_input[1]:
                raise ValueError("Invalid entry: both lists are equal.")
            elif any(x == y for x, y in zip(data_input[0], data_input[1])):
                raise ValueError(
                    "Invalid entry: at least one variable contains the same value for the lower and upper bounds."
                )
            else:
                bounds_array = np.zeros(
                    (
                        2,
                        len(data_input[0]),
                    )
                )
                bounds_array[0, :] = np.array(data_input[0])
                bounds_array[1, :] = np.array(data_input[1])
                data_headers = []
            self.data = bounds_array
            self.x_data = bounds_array
            self.data_headers = data_headers

        if edges is None:
            edges = True
            self.edge = edges
        elif not isinstance(edges, bool):
            raise TypeError('Invalid "edges" entry. Must be boolean')
        elif (edges is True) or (edges is False):
            self.edge = edges

        # Check that list_of_samples_per_variable is a list, list length is correct, all dimensions greater than 1 and all list values are integers
        if not isinstance(list_of_samples_per_variable, list):
            raise TypeError("list_of_samples_per_variable: list required.")
        if len(list_of_samples_per_variable) != self.x_data.shape[1]:
            raise ValueError(
                "Length of list_of_samples_per_variable must equal the number of variables."
            )
        if min(list_of_samples_per_variable) < 2:
            raise ValueError(
                "All variables must have at least two points per dimension"
            )
        if all(isinstance(q, int) for q in list_of_samples_per_variable) is False:
            raise TypeError("All values in list must be integers")

        self.dim_vector = list_of_samples_per_variable
        self.number_of_samples = int(np.prod(self.dim_vector))

        if (
            self.sampling_type == "selection"
            and self.number_of_samples > self.data.shape[0]
        ):
            raise ValueError(
                "Sample size cannot be greater than number of samples in the input data set"
            )

    def sample_points(self):
        """
        ``sample_points`` generates or selects full-factorial designs from an input dataset or data range.

        Returns:
            NumPy Array or Pandas Dataframe:     A numpy array or Pandas dataframe containing the sample points generated or selected by full-factorial sampling.

        """

        points_spread = []
        if self.edge is True:
            for i in self.dim_vector:
                variable_spread = np.arange(i) / (i - 1)
                points_spread.append(variable_spread)
        elif self.edge is False:
            for i in self.dim_vector:
                variable_spread = np.arange(i + 1) / i
                shifted_points = [
                    (variable_spread[i] + variable_spread[i - 1]) / 2
                    for i in range(1, len(variable_spread))
                ]
                points_spread.append(shifted_points)
        samples_list = list(itertools.product(*points_spread))
        samples_array = np.asarray(samples_list)
        unique_sample_points = self.sample_point_selection(
            self.data, samples_array, self.sampling_type
        )
        if len(self.data_headers) > 0 and self.df_flag:
            unique_sample_points = pd.DataFrame(
                unique_sample_points, columns=self.data_headers
            )
        return unique_sample_points


class HaltonSampling(SamplingMethods):
    """
    A class that performs Halton Sampling.

    Halton samples are based on the reversing/flipping the base conversion of numbers using primes.

    To generate n samples in a :math:`p`-dimensional space, the first :math:`p` prime numbers are used to generate the samples.

    Note:
        Use of this method is limited to use in low-dimensionality problems (less than 10 variables). At higher dimensions, the performance of the sampling method has been shown to degrade.

    To use: call class with inputs, and then ``sample_points`` function.

    **Example:**

    .. code-block:: python

        # For the first 10 Halton samples in a 2-D space:
        >>> b = rbf.HaltonSampling(data, 10, sampling_type="selection")
        >>> samples = b.sample_points()

    """

    def __init__(
        self,
        data_input,
        number_of_samples=None,
        sampling_type=None,
        xlabels=None,
        ylabels=None,
    ):
        """

        Initialization of **HaltonSampling** class. Two inputs are required.

        Args:
            data_input (NumPy Array, Pandas Dataframe or list) : The input data set or range to be sampled.

                - When the aim is to select a set of samples from an existing dataset, the dataset must be a NumPy Array or a Pandas Dataframe and **sampling_type** option must be set to "selection". A single output variable (y) is assumed to be supplied in the last column if **xlabels** and **ylabels** are not supplied.
                - When the aim is to generate a set of samples from a data range, the dataset must be a list containing two lists of equal lengths which contain the variable bounds and **sampling_type** option must be set to "creation". It is assumed that the range contains no output variable information  in this case.

            number_of_samples(int): The number of samples to be generated. Should be a positive integer less than or equal to the number of entries (rows) in **data_input**.
            sampling_type(str) : Option which determines whether the algorithm selects samples from an existing dataset ("selection") or attempts to generate sample from a supplied range ("creation"). Default is "creation".

        Keyword Args:
            xlabels (list): List of column names (if **data_input** is a dataframe) or column numbers (if **data_input** is an array) for the independent/input  variables.  Only used in "selection" mode. Default is None.
            ylabels (list): List of column names (if **data_input** is a dataframe) or column numbers (if **data_input** is an array) for the dependent/output variables. Only used in "selection" mode. Default is None.

        Returns:
            **self** function containing the input information.

        Raises:
            ValueError: The input data (**data_input**) is the wrong type/dimension, or **number_of_samples** is invalid (too large, zero, or negative)

            TypeError: When **number_of_samples** is not the right type, or **sampling_type** entry is not a string.

            IndexError: When invalid column names are supplied in **xlabels** or **ylabels**

        """
        if sampling_type is None:
            sampling_type = "creation"
            self.sampling_type = sampling_type
            print("Creation-type sampling will be used.")
        elif not isinstance(sampling_type, str):
            raise TypeError("Invalid sampling type entry. Must be of type <str>.")
        elif (sampling_type.lower() == "creation") or (
            sampling_type.lower() == "selection"
        ):
            sampling_type = sampling_type.lower()
            self.sampling_type = sampling_type
        else:
            raise ValueError(
                'Invalid sampling type requirement entered. Enter "creation" for sampling from a range or "selection" for selecting samples from a dataset.'
            )
        print("Sampling type: ", self.sampling_type, "\n")

        if self.sampling_type == "selection":
            if isinstance(data_input, (pd.DataFrame, np.ndarray)):
                self.selection_columns_preprocessing(data_input, xlabels, ylabels)
            else:
                raise ValueError(
                    'Pandas dataframe or numpy array required for sampling_type "selection."'
                )

            # Catch potential errors in number_of_samples
            if number_of_samples is None:
                print(
                    "\nNo entry for number of samples to be generated. The default value of 5 will be used."
                )
                number_of_samples = 5
            elif number_of_samples > self.data.shape[0]:
                raise ValueError(
                    "Sample size cannot be greater than number of samples in the input data set"
                )
            elif not isinstance(number_of_samples, int):
                raise TypeError("number_of_samples must be an integer.")
            elif number_of_samples <= 0:
                raise ValueError("number_of_samples must a positive, non-zero integer.")
            self.number_of_samples = number_of_samples

        elif self.sampling_type == "creation":
            if not isinstance(data_input, list):
                raise TypeError(
                    'List entry of two elements expected for sampling_type "creation."'
                )
            elif (
                len(data_input) != 2
                or not isinstance(data_input[0], list)
                or not isinstance(data_input[1], list)
                or len(data_input[0]) != len(data_input[1])
            ):
                raise ValueError("data_input must contain two lists of equal lengths.")
            elif data_input[0] == data_input[1]:
                raise ValueError("Invalid entry: both lists are equal.")
            elif any(x == y for x, y in zip(data_input[0], data_input[1])):
                raise ValueError(
                    "Invalid entry: at least one variable contains the same value for the lower and upper bounds."
                )
            else:
                bounds_array = np.zeros(
                    (
                        2,
                        len(data_input[0]),
                    )
                )
                bounds_array[0, :] = np.array(data_input[0])
                bounds_array[1, :] = np.array(data_input[1])
                data_headers = []
            self.data = bounds_array
            self.data_headers = data_headers

            # Catch potential errors in number_of_samples
            if number_of_samples is None:
                print(
                    "\nNo entry for number of samples to be generated. The default value of 5 will be used."
                )
                number_of_samples = 5
            elif not isinstance(number_of_samples, int):
                raise TypeError("number_of_samples must be an integer.")
            elif number_of_samples <= 0:
                raise ValueError("number_of_samples must a positive, non-zero integer.")
            self.number_of_samples = number_of_samples
            self.x_data = bounds_array  # Only x data will be present in this case

        if self.x_data.shape[1] > 10:
            # PYLINT-TODO
            # pylint: disable-next=broad-exception-raised
            raise Exception(
                "Dimensionality problem: This method is not available for problems with dimensionality > 10: the performance of the method degrades substantially at higher dimensions"
            )

    def sample_points(self):
        """
        The ``sample_points`` method generates the Halton samples. The steps followed here are:

            1. Determine the number of features in the input data.
            2. Generate the list of primes to be considered by calling ``prime_number_generator`` from the sampling superclass.
            3. Create the first **number_of_samples** elements of the Halton sequence for each prime.
            4. Create the Halton samples by combining the corresponding elements of the Halton sequences for each prime.
            5. When in "selection" mode, determine the closest corresponding point in the input dataset using Euclidean distance minimization. This is done by calling the ``nearest_neighbours`` method in the sampling superclass.

        Returns:
            NumPy Array or Pandas Dataframe:     A numpy array or Pandas dataframe containing **number_of_samples** Halton sample points.

        """
        no_features = self.x_data.shape[1]
        # Generate list of no_features prime numbers
        prime_list = self.prime_number_generator(no_features)
        sample_points = np.zeros((self.number_of_samples, no_features))
        for i in range(0, no_features):
            sample_points[:, i] = self.data_sequencing(
                self.number_of_samples, prime_list[i]
            )
        # Scale input data, then find data points closest in sample space. Unscale before returning points
        unique_sample_points = self.sample_point_selection(
            self.data, sample_points, self.sampling_type
        )
        if len(self.data_headers) > 0 and self.df_flag:
            unique_sample_points = pd.DataFrame(
                unique_sample_points, columns=self.data_headers
            )
        return unique_sample_points


class HammersleySampling(SamplingMethods):
    r"""
    A class that performs Hammersley Sampling.

    Hammersley samples are generated in a similar way to Halton samples - based on the reversing/flipping the base conversion of numbers using primes.

    To generate :math:`n` samples in a :math:`p`-dimensional space, the first :math:`\left(p-1\right)` prime numbers are used to generate the samples. The first dimension is obtained by uniformly dividing the region into **no_samples points**.

    Note:
        Use of this method is limited to use in low-dimensionality problems (less than 10 variables). At higher dimensionalities, the performance of the sampling method has been shown to degrade.

    To use: call class with inputs, and then ``sample_points`` function.

    **Example:**

    .. code-block:: python

        # For the first 10 Hammersley samples in a 2-D space:
        >>> b = rbf.HammersleySampling(data, 10, sampling_type="selection")
        >>> samples = b.sample_points()

    """

    def __init__(
        self,
        data_input,
        number_of_samples=None,
        sampling_type=None,
        xlabels=None,
        ylabels=None,
    ):
        """
        Initialization of **HammersleySampling** class. Two inputs are required.

        Args:
            data_input (NumPy Array, Pandas Dataframe or list): The input data set or range to be sampled.

               - When the aim is to select a set of samples from an existing dataset, the dataset must be a NumPy Array or a Pandas Dataframe and **sampling_type** option must be set to "selection". A single output variable (y) is assumed to be supplied in the last column if **xlabels** and **ylabels** are not supplied.
               - When the aim is to generate a set of samples from a data range, the dataset must be a list containing two lists of equal lengths which contain the variable bounds and **sampling_type** option must be set to "creation". It is assumed that the range contains no output variable information  in this case.

            number_of_samples(int): The number of samples to be generated. Should be a positive integer less than or equal to the number of entries (rows) in **data_input**.
            sampling_type(str) : Option which determines whether the algorithm selects samples from an existing dataset ("selection") or attempts to generate sample from a supplied range ("creation"). Default is "creation".

        Keyword Args:
            xlabels (list): List of column names (if **data_input** is a dataframe) or column numbers (if **data_input** is an array) for the independent/input  variables.  Only used in "selection" mode. Default is None.
            ylabels (list): List of column names (if **data_input** is a dataframe) or column numbers (if **data_input** is an array) for the dependent/output variables. Only used in "selection" mode. Default is None.

            Returns:
                **self** function containing the input information.

            Raises:
                ValueError: The input data (**data_input**) is the wrong type/dimension, or **number_of_samples** is invalid (too large, zero, or negative)

                TypeError: When **number_of_samples** is not the right type, or **sampling_type** entry is not a string.

                IndexError: When invalid column names are supplied in **xlabels** or **ylabels**

        """
        if sampling_type is None:
            sampling_type = "creation"
            self.sampling_type = sampling_type
            print("Creation-type sampling will be used.")
        elif not isinstance(sampling_type, str):
            raise TypeError("Invalid sampling type entry. Must be of type <str>.")
        elif (sampling_type.lower() == "creation") or (
            sampling_type.lower() == "selection"
        ):
            sampling_type = sampling_type.lower()
            self.sampling_type = sampling_type
        else:
            raise ValueError(
                'Invalid sampling type requirement entered. Enter "creation" for sampling from a range or "selection" for selecting samples from a dataset.'
            )
        print("Sampling type: ", self.sampling_type, "\n")

        if self.sampling_type == "selection":
            if isinstance(data_input, (pd.DataFrame, np.ndarray)):
                self.selection_columns_preprocessing(data_input, xlabels, ylabels)
            else:
                raise ValueError(
                    'Pandas dataframe or numpy array required for sampling_type "selection."'
                )

            # Catch potential errors in number_of_samples
            if number_of_samples is None:
                print(
                    "\nNo entry for number of samples to be generated. The default value of 5 will be used."
                )
                number_of_samples = 5
            elif number_of_samples > self.data.shape[0]:
                raise ValueError(
                    "Sample size cannot be greater than number of samples in the input data set"
                )
            elif not isinstance(number_of_samples, int):
                raise TypeError("number_of_samples must be an integer.")
            elif number_of_samples <= 0:
                raise ValueError("number_of_samples must a positive, non-zero integer.")
            self.number_of_samples = number_of_samples

        elif self.sampling_type == "creation":
            if not isinstance(data_input, list):
                raise ValueError(
                    'List entry of two elements expected for sampling_type "creation."'
                )
            elif (
                len(data_input) != 2
                or not isinstance(data_input[0], list)
                or not isinstance(data_input[1], list)
                or len(data_input[0]) != len(data_input[1])
            ):
                raise ValueError("data_input must contain two lists of equal lengths.")
            elif data_input[0] == data_input[1]:
                raise ValueError("Invalid entry: both lists are equal.")
            elif any(x == y for x, y in zip(data_input[0], data_input[1])):
                raise ValueError(
                    "Invalid entry: at least one variable contains the same value for the lower and upper bounds."
                )
            else:
                bounds_array = np.zeros(
                    (
                        2,
                        len(data_input[0]),
                    )
                )
                bounds_array[0, :] = np.array(data_input[0])
                bounds_array[1, :] = np.array(data_input[1])
                data_headers = []
            self.data = bounds_array
            self.data_headers = data_headers

            # Catch potential errors in number_of_samples
            if number_of_samples is None:
                print(
                    "\nNo entry for number of samples to be generated. The default value of 5 will be used."
                )
                number_of_samples = 5
            elif not isinstance(number_of_samples, int):
                raise TypeError("number_of_samples must be an integer.")
            elif number_of_samples <= 0:
                raise ValueError("number_of_samples must a positive, non-zero integer.")
            self.number_of_samples = number_of_samples
            self.x_data = bounds_array  # Only x data will be present in this case

        if self.x_data.shape[1] > 10:
            # PYLINT-TODO
            # pylint: disable-next=broad-exception-raised
            raise Exception(
                "Dimensionality problem: This method is not available for problems with dimensionality > 10: the performance of the method degrades substantially at higher dimensions"
            )

    def sample_points(self):
        """
        The **sampling_type** method generates the Hammersley sample points. The steps followed here are:

            1. Determine the number of features :math:`n_{f}` in the input data.
            2. Generate the list of :math:`\\left(n_{f}-1\\right)` primes to be considered by calling prime_number_generator.
            3. Divide the space [0,**number_of_samples**-1] into **number_of_samples** places to obtain the first dimension for the Hammersley sequence.
            4. For the other :math:`\\left(n_{f}-1\\right)` dimensions, create first **number_of_samples** elements of the Hammersley sequence for each of the :math:`\\left(n_{f}-1\\right)` primes.
            5. Create the Hammersley samples by combining the corresponding elements of the Hammersley sequences created in steps 3 and 4
            6. When in "selection" mode, determine the closest corresponding point in the input dataset using Euclidean distance minimization. This is done by calling the ``nearest_neighbours`` method in the sampling superclass.

        Returns:
            NumPy Array or Pandas Dataframe:     A numpy array or Pandas dataframe containing **number_of_samples** Hammersley sample points.

        """
        no_features = self.x_data.shape[1]
        if no_features == 1:
            prime_list = []
        else:
            prime_list = self.prime_number_generator(no_features - 1)
        sample_points = np.zeros((self.number_of_samples, no_features))
        sample_points[:, 0] = (
            np.arange(0, self.number_of_samples)
        ) / self.number_of_samples
        for i in range(0, len(prime_list)):  # pylint: disable=consider-using-enumerate
            sample_points[:, i + 1] = self.data_sequencing(
                self.number_of_samples, prime_list[i]
            )

        unique_sample_points = self.sample_point_selection(
            self.data, sample_points, self.sampling_type
        )
        if len(self.data_headers) > 0 and self.df_flag:
            unique_sample_points = pd.DataFrame(
                unique_sample_points, columns=self.data_headers
            )
        return unique_sample_points


class CVTSampling(SamplingMethods):
    """
    A class that constructs Centroidal Voronoi Tessellation (CVT) samples.

    CVT sampling is based on the generation of samples in which the generators of the Voronoi tessellations and the mass centroids coincide.

    To use: call class with inputs, and then ``sample_points`` function.

    **Example:**

    .. code-block:: python

        # For the first 10 CVT samples in a 2-D space:
        >>> b = rbf.CVTSampling(data_bounds, 10, tolerance = 1e-5, sampling_type="creation")
        >>> samples = b.sample_points()

    """

    def __init__(
        self,
        data_input,
        number_of_samples=None,
        tolerance=None,
        sampling_type=None,
        xlabels=None,
        ylabels=None,
        rand_seed=None,
    ):
        """
        Initialization of CVTSampling class. Two inputs are required, while an optional option to control the solution accuracy may be specified.

        Args:
            data_input (NumPy Array, Pandas Dataframe or list): The input data set or range to be sampled.
                - When the aim is to select a set of samples from an existing dataset, the dataset must be a NumPy Array or a Pandas Dataframe and **sampling_type** option must be set to "selection". A single output variable (y) is assumed to be supplied in the last column if **xlabels** and **ylabels** are not supplied.
                - When the aim is to generate a set of samples from a data range, the dataset must be a list containing two lists of equal lengths which contain the variable bounds and **sampling_type** option must be set to "creation". It is assumed that the range contains no output variable information  in this case.


            number_of_samples(int): The number of samples to be generated. Should be a positive integer less than or equal to the number of entries (rows) in **data_input**.
            sampling_type(str) : Option which determines whether the algorithm selects samples from an existing dataset ("selection") or attempts to generate sample from a supplied range ("creation"). Default is "creation".

        Keyword Args:
            xlabels (list): List of column names (if **data_input** is a dataframe) or column numbers (if **data_input** is an array) for the independent/input  variables.  Only used in "selection" mode. Default is None.
            ylabels (list): List of column names (if **data_input** is a dataframe) or column numbers (if **data_input** is an array) for the dependent/output variables. Only used in "selection" mode. Default is None.
            rand_seed (int): Option that allows users to fix the numpy random seed generator for reproducibility (if required).
            tolerance(float): Maximum allowable Euclidean distance between centres from consecutive iterations of the algorithm. Termination condition for algorithm.

                - The smaller the value of tolerance, the better the solution but the longer the algorithm requires to converge. Default value is :math:`10^{-7}`.

        Returns:
                **self** function containing the input information.


        Raises:
                ValueError: The input data (**data_input**) is the wrong type/dimension, or **number_of_samples** is invalid (too large, zero, or negative)

                ValueError: When the tolerance specified is too loose (tolerance > 0.1)

                TypeError: When **number_of_samples** is not the right type, or **sampling_type** entry is not a string

                IndexError: When invalid column names are supplied in **xlabels** or **ylabels**

                Exception: When the tolerance specified is invalid

                warnings.warn: when the tolerance specified by the user is too tight (tolerance < :math:`10^{-9}`)

        """
        if sampling_type is None:
            sampling_type = "creation"
            self.sampling_type = sampling_type
            print("Creation-type sampling will be used.")
        elif not isinstance(sampling_type, str):
            raise TypeError("Invalid sampling type entry. Must be of type <str>.")
        elif (sampling_type.lower() == "creation") or (
            sampling_type.lower() == "selection"
        ):
            sampling_type = sampling_type.lower()
            self.sampling_type = sampling_type
        else:
            raise ValueError(
                'Invalid sampling type requirement entered. Enter "creation" for sampling from a range or "selection" for selecting samples from a dataset.'
            )
        print("Sampling type: ", self.sampling_type, "\n")

        if self.sampling_type == "selection":
            if isinstance(data_input, (pd.DataFrame, np.ndarray)):
                self.selection_columns_preprocessing(data_input, xlabels, ylabels)
            else:
                raise ValueError(
                    'Pandas dataframe or numpy array required for sampling_type "selection."'
                )

            # Catch potential errors in number_of_samples
            if number_of_samples is None:
                print(
                    "\nNo entry for number of samples to be generated. The default value of 5 will be used."
                )
                number_of_samples = 5
            elif number_of_samples > self.data.shape[0]:
                raise ValueError(
                    "CVT sample size cannot be greater than number of samples in the input data set"
                )
            elif not isinstance(number_of_samples, int):
                raise TypeError("number_of_samples must be an integer.")
            elif number_of_samples <= 0:
                raise ValueError("number_of_samples must a positive, non-zero integer.")
            self.number_of_centres = number_of_samples

        elif self.sampling_type == "creation":
            if not isinstance(data_input, list):
                raise ValueError(
                    'List entry of two elements expected for sampling_type "creation."'
                )
            elif (
                len(data_input) != 2
                or not isinstance(data_input[0], list)
                or not isinstance(data_input[1], list)
                or len(data_input[0]) != len(data_input[1])
            ):
                raise ValueError("data_input must contain two lists of equal lengths.")
            elif data_input[0] == data_input[1]:
                raise ValueError("Invalid entry: both lists are equal.")
            elif any(x == y for x, y in zip(data_input[0], data_input[1])):
                raise ValueError(
                    "Invalid entry: at least one variable contains the same value for the lower and upper bounds."
                )
            else:
                bounds_array = np.zeros(
                    (
                        2,
                        len(data_input[0]),
                    )
                )
                bounds_array[0, :] = np.array(data_input[0])
                bounds_array[1, :] = np.array(data_input[1])
                data_headers = []
            self.data = bounds_array
            self.data_headers = data_headers

            # Catch potential errors in number_of_samples
            if number_of_samples is None:
                print(
                    "\nNo entry for number of samples to be generated. The default value of 5 will be used."
                )
                number_of_samples = 5
            elif not isinstance(number_of_samples, int):
                raise TypeError("number_of_samples must be an integer.")
            elif number_of_samples <= 0:
                raise ValueError("number_of_samples must a positive, non-zero integer.")
            self.number_of_centres = number_of_samples

            x_data = bounds_array  # Only x data will be present in this case
            if x_data.ndim == 1:
                x_data = x_data.reshape(len(x_data), 1)
            self.x_data = x_data
            # self.y_data = []

        if tolerance is None:
            tolerance = 1e-7
        elif tolerance > 0.1:
            raise ValueError("Tolerance must be less than 0.1 to achieve good results")
        elif tolerance < 1e-9:
            _log.warning(
                "Tolerance too tight. CVT algorithm may take long time to converge."
            )
        elif (tolerance < 0.1) and (tolerance > 1e-9):
            # valid tolerance
            pass
        else:
            # PYLINT-TODO
            # pylint: disable-next=broad-exception-raised
            raise Exception("Invalid tolerance input")
        self.eps = tolerance

        if rand_seed is not None:
            try:
                self.seed_value = int(rand_seed)
                np.random.seed(self.seed_value)
            except ValueError:
                raise ValueError("Random seed must be an integer.")

    @staticmethod
    def random_sample_selection(no_samples, no_features):
        """
        Function generates a the required number of samples (no_samples) within an no_features-dimensional space.
        This is achieved by generating an m x n 2-D array using numpy's random.rand function, where

            - m = number of training samples to be generated, and'
            - n = number of design features/variables (dimensionality of the problem).

        Args:
            no_samples(int): The number of samples to be generated.
            no_features(int): Number of design features/variables in the input data.

        Returns:
            random_points(NumPy Array): 2-D array of size no_samples x no_features generated from a uniform distribution.

        Example: Generate three samples for a two-dimensional problem
            >>  rbf.CVTSampling.random_sample_selection(3, 2)
            >> array([[0.03149075, 0.70566624], [0.48319597, 0.03810093], [0.19962214, 0.57641408]])

        """
        random_points = np.random.rand(no_samples, no_features)
        return random_points

    @staticmethod
    def eucl_distance(u, v):
        """
        The function eucl_distance(u,v) calculates Euclidean distance between two points or arrays u and v.

        Args:
            u, v (NumPy Array): Two points or arrays with the same number of features (same second dimension)

        Returns:
            euc_d(NumPy Array): Array of size (u.shape[0] x 1) containing Euclidean distances.

        """
        d = u - v
        d_sq = d**2
        euc_d = np.sqrt(np.sum(d_sq, axis=1))
        return euc_d

    @staticmethod
    def create_centres(
        initial_centres, current_random_points, current_centres, counter
    ):
        """
        The function create_centres generates new mass centroids for the design space based on McQueen's method.
        The mass centroids are created based on the previous mass centroids and the mean of random data sampling the design space.

            Args:
                initial_centres(NumPy Array): A 2-D array containing the current mass centroids, size no_samples x no_features.
                current_random_points(NumPy Array): A 2-D array containing several points generated randomly from within the design space.
                current_centres(NumPy Array): Array containing the index number of the closest mass centroid of each point in current_random_points, representing its class.
                counter(int): current iteration number

            Returns:
                centres(NumPy Array): A 2-D array containing the new mass centroids, size no_samples x no_features.

        The steps carried out in the function at each iteration are:
        (1) Classify the current random points in current_random_points based on their centres
        (2) Evaluate the mean of the random points in each class
        (3) Create the new centres as the weighted average of the current centres (initial_centres) and the mean data calculated in the second step. The weighting is done based on the number of iterations (counter).

        """
        centres = np.zeros((initial_centres.shape[0], initial_centres.shape[1]))
        current_centres = current_centres.reshape(current_centres.shape[0], 1)
        for i in range(0, initial_centres.shape[0]):
            data_matrix = current_random_points[current_centres[:, 0] == i]
            m_prime, n_prime = data_matrix.shape  # pylint: disable=unused-variable
            if m_prime == 0:
                centres[i, :] = np.mean(initial_centres, axis=0)
            else:
                centres[i, :] = np.mean(data_matrix, axis=0)

        # Weighted average based on previous number of iterations
        centres = ((counter * initial_centres) + centres) / (counter + 1)
        return centres

    def sample_points(self):
        """
        The ``sample_points`` method determines the best/optimal centre points (centroids) for a data set based on the minimization of the total distance between points and centres.

        Procedure based on McQueen's algorithm: iteratively minimize distance, and re-position centroids.
        Centre re-calculation done as the mean of each data cluster around each centre.

        Returns:
            NumPy Array or Pandas Dataframe:     A numpy array or Pandas dataframe containing the final **number_of_samples** centroids obtained by the CVT algorithm.

        """
        _, n = self.x_data.shape
        size_multiple = 1000
        initial_centres = self.random_sample_selection(self.number_of_centres, n)
        # Iterative optimization process
        cost_old = 0
        cost_new = 0
        cost_change = float("Inf")
        counter = 1
        while (cost_change > self.eps) and (counter <= 1000):
            cost_old = cost_new
            current_random_points = self.random_sample_selection(
                self.number_of_centres * size_multiple, n
            )
            distance_matrix = np.zeros(
                (current_random_points.shape[0], initial_centres.shape[0])
            )  # Vector to store distances from centroids
            current_centres = np.zeros(
                (current_random_points.shape[0], 1)
            )  # Vector containing the centroid each point belongs to

            # Calculate distance between random points and centres, sort and estimate new centres
            for i in range(0, self.number_of_centres):
                distance_matrix[:, i] = self.eucl_distance(
                    current_random_points, initial_centres[i, :]
                )
            current_centres = np.argmin(distance_matrix, axis=1)
            new_centres = self.create_centres(
                initial_centres, current_random_points, current_centres, counter
            )

            # Estimate distance between new and old centres
            distance_btw_centres = self.eucl_distance(new_centres, initial_centres)
            cost_new = np.sqrt(np.sum(distance_btw_centres**2))
            cost_change = np.abs(cost_old - cost_new)
            counter += 1
            # print(counter, cost_change)
            if cost_change >= self.eps:
                initial_centres = new_centres

        sample_points = new_centres

        unique_sample_points = self.sample_point_selection(
            self.data, sample_points, self.sampling_type
        )
        if len(self.data_headers) > 0 and self.df_flag:
            unique_sample_points = pd.DataFrame(
                unique_sample_points, columns=self.data_headers
            )
        return unique_sample_points


class CustomSampling(SamplingMethods):
    """
    A class that performs custom sampling per dimension as specified by the user. The distribution to be applied per dimension must be specified by the user.

    - The distribution to be used per variable needs to be specified in a list.

    - Users are urged to visit the documentation for more information about normal distribution-based sampling.

    To use: call class with inputs, and then ``sample_points`` function

    **Example:**

    .. code-block:: python

        # To select 50 samples drom a dataset:
        >>> b = rbf.CustomSampling(data, 50, list_of_distributions= ['normal', 'uniform'], sampling_type="selection")
        >>> samples = b.sample_points()

    Note:

        To remain consistent with the other sampling methods and distributions, **bounds are required for specifying normal distributions, rather than the mean and standard deviation**.

        Given the mean (:math:`\\bar{x}`) and standard deviation (:math:`\\sigma`), the bounds of the normal distribution may be computed as:

        Lower bound = :math:`\\bar{x} - 3\\sigma` ; Upper bound = :math:`\\bar{x} + 3\\sigma`

        Users should visit the documentation for more information.

    """

    def __init__(
        self,
        data_input,
        number_of_samples=None,
        list_of_distributions=None,
        sampling_type=None,
        xlabels=None,
        ylabels=None,
        strictly_enforce_gaussian_bounds=False,
        rand_seed=None,
    ):
        """
        Initialization of CustomSampling class. Four inputs are required.

        Args:
            data_input (NumPy Array, Pandas Dataframe or list) :  The input data set or range to be sampled.

                - When the aim is to select a set of samples from an existing dataset, the dataset must be a NumPy Array or a Pandas Dataframe and **sampling_type** option must be set to "selection". A single output variable (y) is assumed to be supplied in the last column if **xlabels** and **ylabels** are not supplied.
                - When the aim is to generate a set of samples from a data range, the dataset must be a list containing two lists of equal lengths which contain the variable bounds and **sampling_type** option must be set to "creation". It is assumed that the range contains no output variable information in this case.

            number_of_samples(int): The number of samples to be generated. Should be a positive integer less than or equal to the number of entries (rows) in **data_input**.
            list_of_distributions (list): The list containing the probability distribution for each variable. The length of the list must match the number of input (i.e. dependent) variables to be sampled. We currently support random, uniform and normal (i.e. Gaussian) distributions.
            sampling_type (str) : Option which determines whether the algorithm selects samples from an existing dataset ("selection") or attempts to generate sample from a supplied range ("creation"). Default is "creation".

        Keyword Args:
            xlabels (list): List of column names (if **data_input** is a dataframe) or column numbers (if **data_input** is an array) for the independent/input  variables.  Only used in "selection" mode. Default is None.
            ylabels (list): List of column names (if **data_input** is a dataframe) or column numbers (if **data_input** is an array) for the dependent/output variables. Only used in "selection" mode. Default is None.
            rand_seed (int): Option that allows users to fix the numpy random seed generator for reproducibility (if required).
            strictly_enforce_gaussian_bounds (bool): Boolean specifying whether the provided bounds for normal distributions should be strictly enforced. Note that selecting this option may affect the underlying distribution. Default is False.

        Returns:
            **self** function containing the input information

        Raises:
            ValueError: The input data (**data_input**) is the wrong type/dimension, or **number_of_samples** is invalid (too large, zero, or negative), **list_of_distributions** is the wrong length, or a non-implemented distribution is supplied in **list_of_distributions**.

            TypeError: When **number_of_samples** is not an integer, **list_of_distributions** is not a list, or **sampling_type** entry is not a string

            IndexError: When invalid column names are supplied in **xlabels** or **ylabels**


        """
        if sampling_type is None:
            sampling_type = "creation"
            self.sampling_type = sampling_type
            print("Creation-type sampling will be used.")
        elif not isinstance(sampling_type, str):
            raise TypeError("Invalid sampling type entry. Must be of type <str>.")
        elif (sampling_type.lower() == "creation") or (
            sampling_type.lower() == "selection"
        ):
            sampling_type = sampling_type.lower()
            self.sampling_type = sampling_type
        else:
            raise ValueError(
                'Invalid sampling type requirement entered. Enter "creation" for sampling from a range or "selection" for selecting samples from a dataset.'
            )
        print("Sampling type: ", self.sampling_type, "\n")

        if self.sampling_type == "selection":
            if isinstance(data_input, (pd.DataFrame, np.ndarray)):
                self.selection_columns_preprocessing(data_input, xlabels, ylabels)
            else:
                raise ValueError(
                    'Pandas dataframe or numpy array required for sampling_type "selection."'
                )

            # Catch potential errors in number_of_samples
            if number_of_samples is None:
                print(
                    "\nNo entry for number of samples to be generated. The default value of 5 will be used."
                )
                number_of_samples = 5
            elif number_of_samples > self.data.shape[0]:
                raise ValueError(
                    "Sample size cannot be greater than number of samples in the input data set"
                )
            elif not isinstance(number_of_samples, int):
                raise TypeError("number_of_samples must be an integer.")
            elif number_of_samples <= 0:
                raise ValueError("number_of_samples must a positive, non-zero integer.")
            self.number_of_samples = number_of_samples

        elif self.sampling_type == "creation":
            if not isinstance(data_input, list):
                raise ValueError(
                    'List entry of two elements expected for sampling_type "creation."'
                )
            elif (
                len(data_input) != 2
                or not isinstance(data_input[0], list)
                or not isinstance(data_input[1], list)
                or len(data_input[0]) != len(data_input[1])
            ):
                raise ValueError("data_input must contain two lists of equal lengths.")
            elif data_input[0] == data_input[1]:
                raise ValueError("Invalid entry: both lists are equal.")
            elif any(x == y for x, y in zip(data_input[0], data_input[1])):
                raise ValueError(
                    "Invalid entry: at least one variable contains the same value for the lower and upper bounds."
                )
            else:
                bounds_array = np.zeros(
                    (
                        2,
                        len(data_input[0]),
                    )
                )
                bounds_array[0, :] = np.array(data_input[0])
                bounds_array[1, :] = np.array(data_input[1])
                data_headers = []
            self.data = bounds_array
            self.data_headers = data_headers

            # Catch potential errors in number_of_samples
            if number_of_samples is None:
                print(
                    "\nNo entry for number of samples to be generated. The default value of 5 will be used."
                )
                number_of_samples = 5
            elif not isinstance(number_of_samples, int):
                raise TypeError("number_of_samples must be an integer.")
            elif number_of_samples <= 0:
                raise ValueError("number_of_samples must a positive, non-zero integer.")
            self.number_of_samples = number_of_samples
            self.x_data = bounds_array  # Only x data will be present in this case

        # Check that list_of_distributions is a list, list length is correct and all list values are strings
        if list_of_distributions is None:
            raise ValueError("list_of_distributions cannot be empty.")
        if not isinstance(list_of_distributions, list):
            raise TypeError("Error with list_of_distributions: list required.")
        if len(list_of_distributions) != self.x_data.shape[1]:
            raise ValueError(
                "Length of list_of_distributions must equal the number of variables."
            )
        if all(isinstance(q, str) for q in list_of_distributions) is False:
            raise TypeError("All values in list must be strings")
        if not all(
            q.lower() in ["random", "normal", "uniform"] for q in list_of_distributions
        ):
            raise ValueError(
                "list_of_distributions only supports 'random', 'normal' and 'uniform' sampling options."
            )
        self.dist_vector = list_of_distributions

        if not isinstance(strictly_enforce_gaussian_bounds, bool):
            raise TypeError(
                "Invalid 'strictly_enforce_gaussian_bounds' entry. Must be boolean."
            )
        self.normal_bounds_enforced = strictly_enforce_gaussian_bounds

        if rand_seed is not None:
            try:
                self.seed_value = int(rand_seed)
            except ValueError:
                raise ValueError("Random seed must be an integer.")
        else:
            self.seed_value = rand_seed

    def generate_from_dist(self, dist_name):
        if dist_name.lower() in ["uniform", "random"]:
            dist = getattr(np.random.default_rng(self.seed_value), dist_name.lower())
            var_values = np.array(dist(size=self.number_of_samples))
            return dist, var_values
        elif dist_name.lower() == "normal":
            dist = getattr(np.random.default_rng(self.seed_value), "normal")
            var_values = dist(loc=0.5, scale=1 / 6, size=self.number_of_samples)
            if not self.normal_bounds_enforced:
                return dist, np.array(var_values)
            else:
                if (
                    sum([1 for i in range(0, var_values.shape[0]) if var_values[i] > 1])
                    + sum(
                        [1 for i in range(0, var_values.shape[0]) if var_values[i] < 0]
                    )
                    > 0
                ):
                    _log.warning(
                        "Points adjusted to remain within specified Gaussian bounds. This may affect the underlying distribution."
                    )
                    out_locations = [
                        i
                        for i in range(0, var_values.shape[0])
                        if var_values[i] > 1 or var_values[i] < 0
                    ]
                    for k in out_locations:
                        rep_value = var_values[k]
                        while (rep_value < 0) or (rep_value > 1):
                            rep_value = dist(loc=0.5, scale=1 / 6, size=1)
                        var_values[k] = rep_value[0]
                assert (
                    sum([1 for i in range(0, var_values.shape[0]) if var_values[i] > 1])
                    + sum(
                        [1 for i in range(0, var_values.shape[0]) if var_values[i] < 0]
                    )
                    == 0
                )
                return dist, np.array(var_values)

    def sample_points(self):
        points_spread = []
        for i in self.dist_vector:
            _, var_values = self.generate_from_dist(i)
            points_spread.append(var_values)
        samples_array = np.asarray(points_spread).T

        # Scale input data, then find data points closest in sample space. Unscale before returning points
        unique_sample_points = self.sample_point_selection(
            self.data, samples_array, self.sampling_type
        )
        if len(self.data_headers) > 0 and self.df_flag:
            unique_sample_points = pd.DataFrame(
                unique_sample_points, columns=self.data_headers
            )
        return unique_sample_points



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Creating a Component derived from Pyomo's Expression to use in cases
where an Expression could be mistaken for a Var.
"""
# TODO: Missing docstrings
# pylint: disable=missing-class-docstring
# pylint: disable=missing-function-docstring

import pyomo.environ as pyo
from pyomo.core.base.expression import ExpressionData
from pyomo.core.base.component import ModelComponentFactory
from pyomo.core.base.indexed_component import (
    UnindexedComponent_set,
)
from pyomo.core.base.disable_methods import disable_methods


# Author: Andrew Lee
class VarLikeExpressionData(ExpressionData):
    """
    An object derived from ExpressionData which implements methods for
    common APIs on Vars.

    Constructor Arguments:
        expr: The Pyomo expression stored in this expression.

        component: The Expression object that owns this data.

    Public Class Attributes:
        expr: The expression owned by this data.

    Private class attributes:
        _component: The expression component.
    """

    # Define methods for common APIs on Vars in case user mistakes
    # an Expression for a Var
    def set_value(self, value, force=False):
        """
        Overload set_value method to provide meaningful error if user attempts
        to set the value of the Expression. In order to support changing the
        expression (and setting it originally), if self.expr is None or
        force=True, the value of the expression will be updated, otherwise a
        TypeError will be raised.

        Args:
            value: value to set for expr
            force: force updating of expr if True (default = False)

        Returns:
            None

        Raises:
            TypeError if expr is not None and force=False
        """
        if self.arg(0) is None or force:
            super().set_value(value)
        else:
            raise TypeError(
                f"{self.name} is an Expression and does not have a value "
                f"which can be set."
            )

    @property
    def value(self):
        raise TypeError(
            f"{self.name} is an Expression and does not have a value "
            f"attribute. Use the 'value()' method instead."
        )

    @value.setter
    def value(self, expr):
        raise TypeError(
            "%s is an Expression and does not have a value which can be set."
            % (self.name)
        )

    def setlb(self, val=None):
        raise TypeError(
            "%s is an Expression and can not have bounds. "
            "Use an inequality Constraint instead." % (self.name)
        )

    def setub(self, val=None):
        raise TypeError(
            "%s is an Expression and can not have bounds. "
            "Use an inequality Constraint instead." % (self.name)
        )

    def fix(self, val=None):
        raise TypeError(
            "%s is an Expression and can not be fixed. "
            "Use an equality Constraint instead." % (self.name)
        )

    def unfix(self):
        raise TypeError("%s is an Expression and can not be unfixed." % (self.name))


@ModelComponentFactory.register(
    "Named expressions that can be used in places of variables."
)
class VarLikeExpression(pyo.Expression):
    """
    A shared var-like expression container, which may be defined over a index.

    Constructor Arguments:
        initialize: A Pyomo expression or dictionary of expressions used
        to initialize this object.

        expr: A synonym for initialize.

        rule: A rule function used to initialize this object.
    """

    _ComponentDataClass = VarLikeExpressionData
    NoConstraint = (1000,)
    Skip = (1000,)

    def __new__(cls, *args, **kwds):
        if cls is not VarLikeExpression:
            return super(VarLikeExpression, cls).__new__(cls)
        if not args or (args[0] is UnindexedComponent_set and len(args) == 1):
            return super(VarLikeExpression, cls).__new__(
                AbstractSimpleVarLikeExpression
            )
        else:
            return super(VarLikeExpression, cls).__new__(IndexedVarLikeExpression)


class SimpleVarLikeExpression(VarLikeExpressionData, VarLikeExpression):
    def __init__(self, *args, **kwds):
        VarLikeExpressionData.__init__(self, expr=None, component=self)
        VarLikeExpression.__init__(self, *args, **kwds)

    #
    # From Pyomo: Leaving this method for backward compatibility reasons.
    # (probably should be removed)
    # Note: Doesn't seem to work without it
    #
    def add(self, index, expr):
        """Add an expression with a given index."""
        if index is not None:
            raise KeyError(
                "SimpleExpression object '%s' does not accept "
                "index values other than None. Invalid value: %s" % (self.name, index)
            )
        if isinstance(expr, tuple) and expr == pyo.Expression.Skip:
            raise ValueError(
                "Expression.Skip can not be assigned "
                "to an Expression that is not indexed: %s" % (self.name)
            )
        self.set_value(expr)
        return self


@disable_methods({"set_value", "is_constant", "is_fixed", "expr"})
class AbstractSimpleVarLikeExpression(SimpleVarLikeExpression):
    pass


class IndexedVarLikeExpression(VarLikeExpression):

    #
    # From Pyomo: Leaving this method for backward compatibility reasons
    # Note: It allows adding members outside of self._index.
    #       This has always been the case. Not sure there is
    #       any reason to maintain a reference to a separate
    #       index set if we allow this.
    #
    def add(self, index, expr):
        """Add an expression with a given index."""
        if isinstance(expr, tuple) and expr == pyo.Expression.Skip:
            return None
        cdata = VarLikeExpressionData(expr, component=self)
        self._data[index] = cdata
        return cdata

    # Define methods for common APIs on Vars in case user mistakes
    # an Expression for a Var
    def setlb(self, val=None):
        raise TypeError(
            "%s is an Expression and can not have bounds. "
            "Use inequality Constraints instead." % (self.name)
        )

    def setub(self, val=None):
        raise TypeError(
            "%s is an Expression and can not have bounds. "
            "Use inequality Constraints instead." % (self.name)
        )

    def fix(self, val=None):
        raise TypeError(
            "%s is an Expression and can not be fixed. "
            "Use equality Constraints instead." % (self.name)
        )

    def unfix(self):
        raise TypeError("%s is an Expression and can not be unfixed." % (self.name))



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
import pytest
import pyomo.environ as pyo
from idaes.apps.grid_integration.tracker import Tracker
from idaes.apps.grid_integration.tests.util import ExampleModel, testing_model_data


class MissingModel:
    """
    A class for testing missing methods and attributes.
    """

    method_list = [
        "populate_model",
        "get_implemented_profile",
        "update_model",
        "get_last_delivered_power",
        "record_results",
        "write_results",
    ]
    method_dict = {k: lambda: None for k in method_list}

    attr_dict = {"power_output": "power_output", "total_cost": ("tot_cost", 1)}

    def __init__(self, missing_method=None, missing_attr=None):
        """
        Constructs a model class without the specified missing methods and/or
        missing attributes.
        """

        for m in self.method_dict:
            if m != missing_method:
                setattr(self, m, self.method_dict[m])

        for a in self.attr_dict:
            if a != missing_attr:
                setattr(self, a, self.attr_dict[a])


horizon = 4


@pytest.mark.unit
def test_model_object_missing_methods():
    n_tracking_hour = 1
    solver = pyo.SolverFactory("cbc")

    # By definition, the model object should contain these methods
    method_list = [
        "populate_model",
        "get_implemented_profile",
        "update_model",
        "get_last_delivered_power",
        "record_results",
        "write_results",
    ]

    # test if the correct error message is raised if a model misses necessary methods
    for m in method_list:
        tracking_model_object = MissingModel(missing_method=m)
        with pytest.raises(AttributeError, match=r".*{}().*".format(m)):
            tracker_object = Tracker(
                tracking_model_object=tracking_model_object,
                tracking_horizon=horizon,
                n_tracking_hour=n_tracking_hour,
                solver=solver,
            )


@pytest.mark.unit
def test_model_object_missing_attr():
    n_tracking_hour = 1
    solver = pyo.SolverFactory("cbc")
    # By definition, the model object should contain these attributes
    attr_list = ["power_output", "total_cost"]

    # test if the correct error message is raised if a model misses necessary attributes
    for a in attr_list:
        tracking_model_object = MissingModel(missing_attr=a)
        with pytest.raises(AttributeError, match=r".*{}.*".format(a)):
            tracker_object = Tracker(
                tracking_model_object=tracking_model_object,
                tracking_horizon=horizon,
                n_tracking_hour=n_tracking_hour,
                solver=solver,
            )


@pytest.mark.unit
def test_n_tracking_hour_checker():
    solver = pyo.SolverFactory("cbc")
    tracking_model_object = ExampleModel(model_data=testing_model_data)

    # test if tracker raise error when negative number of hours is given
    n_tracking_hour = -1
    with pytest.raises(ValueError, match=r".*greater than zero.*"):
        tracker_object = Tracker(
            tracking_model_object=tracking_model_object,
            tracking_horizon=horizon,
            n_tracking_hour=n_tracking_hour,
            solver=solver,
        )

    # test if tracker raise error when floating number of hours is given
    n_tracking_hour = 3.0
    with pytest.raises(TypeError, match=r".*should be an integer.*"):
        tracker_object = Tracker(
            tracking_model_object=tracking_model_object,
            tracking_horizon=horizon,
            n_tracking_hour=n_tracking_hour,
            solver=solver,
        )


@pytest.mark.unit
def test_solver_checker():
    n_tracking_hour = 1
    tracking_model_object = ExampleModel(model_data=testing_model_data)

    # test if bidder raise error when invalid solver is provided
    invalid_solvers = [5, "cbc", "ipopt"]
    for s in invalid_solvers:
        with pytest.raises(TypeError, match=r".*not a valid Pyomo solver.*"):
            tracker_object = Tracker(
                tracking_model_object=tracking_model_object,
                tracking_horizon=horizon,
                n_tracking_hour=n_tracking_hour,
                solver=s,
            )


@pytest.fixture
def tracker_object():
    n_tracking_hour = 1
    solver = pyo.SolverFactory("cbc")

    # create a tracker model
    tracking_model_object = ExampleModel(model_data=testing_model_data)
    tracker_object = Tracker(
        tracking_model_object=tracking_model_object,
        tracking_horizon=horizon,
        n_tracking_hour=n_tracking_hour,
        solver=solver,
    )
    return tracker_object


@pytest.mark.component
@pytest.mark.skipif(
    not pyo.SolverFactory("cbc").available(False), reason="solver not available"
)
def test_track_market_dispatch(tracker_object):
    market_dispatch = [30, 40, 50, 70]
    tracker_object.track_market_dispatch(
        market_dispatch=market_dispatch, date="2021-07-26", hour="17:00"
    )

    for t, dispatch in zip(range(horizon), market_dispatch):
        assert (
            pytest.approx(pyo.value(tracker_object.power_output[t]), abs=1e-3)
            == dispatch
        )

    last_delivered_power = market_dispatch[0]
    assert (
        pytest.approx(tracker_object.get_last_delivered_power(), abs=1e-3)
        == last_delivered_power
    )


@pytest.mark.component
def test_track_deviation_penalty(tracker_object):
    large_penalty = 10000

    assert pytest.approx(large_penalty) == pyo.value(
        tracker_object.model.deviation_penalty[0]
    )

    for t in range(1, horizon):
        assert pytest.approx(
            large_penalty / (horizon - tracker_object.n_tracking_hour)
        ) == pyo.value(tracker_object.model.deviation_penalty[t])



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Simple model to represent resistance originating from the contact between the
flow mesh and electrode or flow mesh and interconnect. An equality constraint is
created to add a resistance heating term from the heat flux coming from one side
to that coming out the other.

Boundary variables:
    - ``temperature_deviation[t, iz]``
    - ``heat_flux_x0[t, iz]``
    - ``heat_flux_x1[t, iz]``

Instances of ``Var`` that must be fixed:
    - ``log_preexponential_factor``: Natural logarithm of area squared resistance preexponential factor in ohm * m**2
    - ``thermal_exponent_dividend``: Parameter divided by temperature in area squared resistance equation, in K.
      Would be something like (reduced) activation energy, but it can be both negative and positive.
    - ``contact_fraction``: Fraction of area at which both surfaces touch. If unknown, can fix at one.
"""
# TODO: Missing docstrings
# pylint: disable=missing-class-docstring
# pylint: disable=missing-function-docstring

# TODO: Look into protected access issues
# pylint: disable=protected-access

__author__ = "Douglas Allan"

from pyomo.common.config import ConfigBlock, ConfigValue, In, Bool
import pyomo.environ as pyo


from idaes.core import declare_process_block_class, UnitModelBlockData
import idaes.models_extra.power_generation.unit_models.soc_submodels.common as common
import idaes.core.util.scaling as iscale
from idaes.core.solvers import get_solver

import idaes.logger as idaeslog


@declare_process_block_class("SocContactResistor")
class SocContactResistorData(UnitModelBlockData):
    CONFIG = ConfigBlock()
    CONFIG.declare(
        "dynamic",
        ConfigValue(
            domain=In([False]),
            default=False,
            description="Dynamic model flag",
            doc="No capacities or holdups, so no internal dynamics",
        ),
    )
    CONFIG.declare(
        "has_holdup",
        ConfigValue(domain=In([False]), default=False),
    )
    CONFIG.declare(
        "voltage_drop_custom",
        ConfigValue(
            domain=Bool,
            default=False,
            description="If True, add voltage_drop_custom Var to be connected to degradation models",
        ),
    )
    common._submodel_boilerplate_config(CONFIG)
    common._thermal_boundary_conditions_config(CONFIG, thin=True)

    def build(self):
        super().build()

        # Set up some sets for the space and time indexing
        tset = self.flowsheet().config.time
        # Set up node and face sets and get integer indices for them
        izfaces, iznodes = common._face_initializer(  # pylint: disable=unused-variable
            self, self.config.control_volume_zfaces, "z"
        )
        common._submodel_boilerplate_create_if_none(self)
        common._create_thermal_boundary_conditions_if_none(self, thin=True)

        # Preexponential factor needs to be given in units of ohm*m**2
        self.log_preexponential_factor = pyo.Var(
            initialize=-50, units=pyo.units.dimensionless
        )
        self.thermal_exponent_dividend = pyo.Var(initialize=0, units=pyo.units.K)
        self.contact_fraction = pyo.Var(
            initialize=1, units=pyo.units.dimensionless, bounds=(0, 1)
        )
        if self.config.voltage_drop_custom:
            self.voltage_drop_custom = pyo.Var(
                tset,
                iznodes,
                units=pyo.units.volts,
                doc="Custom voltage drop term for degradation modeling",
            )

        @self.Expression(tset, iznodes)
        def contact_resistance(b, t, iz):
            return (
                1
                * pyo.units.ohm
                * pyo.units.m**2
                / b.contact_fraction
                * pyo.exp(
                    b.log_preexponential_factor
                    + b.thermal_exponent_dividend / (b.temperature[t, iz])
                )
            )

        @self.Expression(tset, iznodes)
        def voltage_drop_total(b, t, iz):
            if self.config.voltage_drop_custom:
                return (
                    b.contact_resistance[t, iz] * b.current_density[t, iz]
                    + b.voltage_drop_custom[t, iz]
                )
            else:
                return b.contact_resistance[t, iz] * b.current_density[t, iz]

        @self.Expression(tset, iznodes)
        def joule_heating_flux(b, t, iz):
            return b.voltage_drop_total[t, iz] * b.current_density[t, iz]

        @self.Constraint(tset, iznodes)
        def heat_flux_x_eqn(b, t, iz):
            return (
                b.heat_flux_x1[t, iz]
                == b.heat_flux_x0[t, iz] + b.joule_heating_flux[t, iz]
            )

    def initialize_build(
        self,
        outlvl=idaeslog.NOTSET,
        solver=None,
        optarg=None,
        fix_heat_flux_x0=True,
        temperature_guess=None,
    ):
        solve_log = idaeslog.getSolveLogger(self.name, outlvl, tag="unit")

        if temperature_guess is not None:
            for t in self.flowsheet().time:
                for iz in self.iznodes:
                    common._set_if_unfixed(self.temperature_z[t, iz], temperature_guess)

        self.temperature_deviation_x.fix()
        if fix_heat_flux_x0:
            self.heat_flux_x0.fix()
        else:
            self.heat_flux_x1.fix()

        solver_obj = get_solver(solver, optarg)
        common._init_solve_block(self, solver_obj, solve_log)

        self.temperature_deviation_x.unfix()
        if fix_heat_flux_x0:
            self.heat_flux_x0.unfix()
        else:
            self.heat_flux_x1.unfix()

    def calculate_scaling_factors(self):
        pass

    def recursive_scaling(self):
        gsf = iscale.get_scaling_factor
        for i, c in self.heat_flux_x_eqn.items():
            if self.heat_flux_x0[i].is_reference():
                sq0 = gsf(self.heat_flux_x0[i].referent)
            else:
                sq0 = gsf(self.heat_flux_x0[i], warning=True)
            if self.heat_flux_x1[i].is_reference():
                sq1 = gsf(self.heat_flux_x1[i].referent)
            else:
                sq1 = gsf(self.heat_flux_x1[i], warning=True)
            sq = min(sq0, sq1)
            iscale.constraint_scaling_transform(c, sq, overwrite=False)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for CLC solid phase thermo state block; tests for construction and solve
Author: Chinedu Okoli
"""

import pytest

from pyomo.environ import check_optimal_termination, ConcreteModel, Var
from pyomo.util.check_units import assert_units_consistent

from idaes.core import FlowsheetBlock

from idaes.core.util.model_statistics import (
    degrees_of_freedom,
    fixed_variables_set,
    activated_constraints_set,
)
from idaes.core.util import scaling as iscale
from idaes.core.solvers import get_solver

from idaes.models_extra.gas_solid_contactors.properties.methane_iron_OC_reduction.solid_phase_thermo import (
    SolidPhaseParameterBlock,
)

# Get default solver for testing
solver = get_solver()


# -----------------------------------------------------------------------------
@pytest.fixture(scope="class")
def solid_prop():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    # solid properties and state inlet block
    m.fs.properties = SolidPhaseParameterBlock()
    m.fs.unit = m.fs.properties.build_state_block(
        [0], parameters=m.fs.properties, defined_state=True
    )

    m.fs.unit[0].flow_mass.fix(1)
    m.fs.unit[0].particle_porosity.fix(0.27)
    m.fs.unit[0].temperature.fix(1183.15)
    m.fs.unit[0].mass_frac_comp["Fe2O3"].fix(0.45)
    m.fs.unit[0].mass_frac_comp["Fe3O4"].fix(1e-9)
    m.fs.unit[0].mass_frac_comp["Al2O3"].fix(0.55)

    return m


@pytest.mark.unit
def test_build_inlet_state_block(solid_prop):
    assert isinstance(solid_prop.fs.unit[0].dens_mass_skeletal, Var)
    assert isinstance(solid_prop.fs.unit[0].enth_mol_comp, Var)
    assert isinstance(solid_prop.fs.unit[0].enth_mass, Var)
    assert isinstance(solid_prop.fs.unit[0].cp_mol_comp, Var)
    assert isinstance(solid_prop.fs.unit[0].cp_mass, Var)


@pytest.mark.unit
def test_setInputs_state_block(solid_prop):
    assert degrees_of_freedom(solid_prop.fs.unit[0]) == 0


@pytest.fixture(scope="class")
def solid_prop_unscaled():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    # solid properties and state inlet block
    m.fs.properties = SolidPhaseParameterBlock()
    m.fs.unit = m.fs.properties.build_state_block(
        [0], parameters=m.fs.properties, defined_state=True
    )

    m.fs.unit[0].flow_mass.fix(1)
    m.fs.unit[0].particle_porosity.fix(0.27)
    m.fs.unit[0].temperature.fix(1183.15)
    m.fs.unit[0].mass_frac_comp["Fe2O3"].fix(0.45)
    m.fs.unit[0].mass_frac_comp["Fe3O4"].fix(1e-9)
    m.fs.unit[0].mass_frac_comp["Al2O3"].fix(0.55)

    return m


@pytest.mark.solver
@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_initialize_unscaled(solid_prop_unscaled):
    orig_fixed_vars = fixed_variables_set(solid_prop_unscaled)
    orig_act_consts = activated_constraints_set(solid_prop_unscaled)

    solid_prop_unscaled.fs.unit.initialize()

    assert degrees_of_freedom(solid_prop_unscaled) == 0

    fin_fixed_vars = fixed_variables_set(solid_prop_unscaled)
    fin_act_consts = activated_constraints_set(solid_prop_unscaled)

    assert len(fin_act_consts) == len(orig_act_consts)
    assert len(fin_fixed_vars) == len(orig_fixed_vars)

    for c in fin_act_consts:
        assert c in orig_act_consts
    for v in fin_fixed_vars:
        assert v in orig_fixed_vars


@pytest.mark.solver
@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_solve_unscaled(solid_prop_unscaled):

    assert hasattr(solid_prop_unscaled.fs.unit[0], "dens_mass_skeletal")
    assert hasattr(solid_prop_unscaled.fs.unit[0], "cp_mass")
    assert hasattr(solid_prop_unscaled.fs.unit[0], "enth_mass")

    results = solver.solve(solid_prop_unscaled)

    # Check for optimal solution
    assert check_optimal_termination(results)


@pytest.mark.component
def test_scaling(solid_prop):
    # Calculate scaling factors

    # Construct property methods to build the constraints

    assert hasattr(solid_prop.fs.unit[0], "dens_mass_skeletal")
    assert hasattr(solid_prop.fs.unit[0], "dens_mass_particle")
    assert hasattr(solid_prop.fs.unit[0], "cp_mol_comp")
    assert hasattr(solid_prop.fs.unit[0], "cp_mass")
    assert hasattr(solid_prop.fs.unit[0], "enth_mass")
    assert hasattr(solid_prop.fs.unit[0], "enth_mol_comp")

    # Call flow and density methods to construct flow and density expressions
    for i in solid_prop.fs.unit[0]._params.component_list:
        solid_prop.fs.unit[0].get_material_flow_terms("Sol", i)
        solid_prop.fs.unit[0].get_material_density_terms("Sol", i)
    solid_prop.fs.unit[0].get_enthalpy_flow_terms("Sol")
    solid_prop.fs.unit[0].get_energy_density_terms("Sol")

    # Calculate scaling factors now that constraints/expressions are built
    iscale.calculate_scaling_factors(solid_prop)

    # Test scaling
    assert pytest.approx(1e-3, abs=1e-2) == iscale.get_scaling_factor(
        solid_prop.fs.unit[0].dens_mass_particle
    )

    for i, c in solid_prop.fs.unit[0].material_flow_terms.items():
        assert pytest.approx(1e-2, abs=1e-2) == iscale.get_scaling_factor(c)
    for i, c in solid_prop.fs.unit[0].material_density_terms.items():
        assert pytest.approx(1e-1, abs=1e-2) == iscale.get_scaling_factor(c)
    for i, c in solid_prop.fs.unit[0].energy_density_terms.items():
        assert pytest.approx(1e-6, abs=1e-2) == iscale.get_scaling_factor(c)
    for i, c in solid_prop.fs.unit[0].enthalpy_flow_terms.items():
        assert pytest.approx(1e-9, abs=1e-2) == iscale.get_scaling_factor(c)

    assert pytest.approx(
        1e-2, abs=1e-2
    ) == iscale.get_constraint_transform_applied_scaling_factor(
        solid_prop.fs.unit[0].density_particle_constraint
    )
    assert pytest.approx(
        1e-5, abs=1e-2
    ) == iscale.get_constraint_transform_applied_scaling_factor(
        solid_prop.fs.unit[0].density_skeletal_constraint
    )
    for i, c in solid_prop.fs.unit[0].cp_shomate_eqn.items():
        assert pytest.approx(
            1e-6, abs=1e-2
        ) == iscale.get_constraint_transform_applied_scaling_factor(c)
    assert pytest.approx(
        1e-6, abs=1e-2
    ) == iscale.get_constraint_transform_applied_scaling_factor(
        solid_prop.fs.unit[0].mixture_heat_capacity_eqn
    )
    for i, c in solid_prop.fs.unit[0].enthalpy_shomate_eqn.items():
        assert pytest.approx(
            1e-6, abs=1e-2
        ) == iscale.get_constraint_transform_applied_scaling_factor(c)
    assert pytest.approx(
        1e-6, abs=1e-2
    ) == iscale.get_constraint_transform_applied_scaling_factor(
        solid_prop.fs.unit[0].mixture_enthalpy_eqn
    )


@pytest.mark.solver
@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_initialize(solid_prop):
    orig_fixed_vars = fixed_variables_set(solid_prop)
    orig_act_consts = activated_constraints_set(solid_prop)

    solid_prop.fs.unit.initialize()

    assert degrees_of_freedom(solid_prop) == 0

    fin_fixed_vars = fixed_variables_set(solid_prop)
    fin_act_consts = activated_constraints_set(solid_prop)

    assert len(fin_act_consts) == len(orig_act_consts)
    assert len(fin_fixed_vars) == len(orig_fixed_vars)

    for c in fin_act_consts:
        assert c in orig_act_consts
    for v in fin_fixed_vars:
        assert v in orig_fixed_vars


@pytest.mark.solver
@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_solve(solid_prop):

    assert hasattr(solid_prop.fs.unit[0], "dens_mass_skeletal")
    assert hasattr(solid_prop.fs.unit[0], "cp_mass")
    assert hasattr(solid_prop.fs.unit[0], "enth_mass")

    results = solver.solve(solid_prop)

    # Check for optimal solution
    assert check_optimal_termination(results)


@pytest.mark.solver
@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_solution(solid_prop):
    assert (
        pytest.approx(3251.75, abs=1e-2)
        == solid_prop.fs.unit[0].dens_mass_skeletal.value
    )
    assert pytest.approx(1, abs=1e-2) == solid_prop.fs.unit[0].cp_mass.value
    assert pytest.approx(0.0039, abs=1e-2) == solid_prop.fs.unit[0].enth_mass.value


@pytest.mark.component
def test_units_consistent(solid_prop):

    # Construct property methods to build the constraints

    assert hasattr(solid_prop.fs.unit[0], "dens_mass_skeletal")
    assert hasattr(solid_prop.fs.unit[0], "dens_mass_particle")
    assert hasattr(solid_prop.fs.unit[0], "cp_mol_comp")
    assert hasattr(solid_prop.fs.unit[0], "cp_mass")
    assert hasattr(solid_prop.fs.unit[0], "enth_mass")
    assert hasattr(solid_prop.fs.unit[0], "enth_mol_comp")

    # Call flow and density methods to construct flow and density expressions
    for i in solid_prop.fs.unit[0]._params.component_list:
        solid_prop.fs.unit[0].get_material_flow_terms("Sol", i)
        solid_prop.fs.unit[0].get_material_density_terms("Sol", i)
    solid_prop.fs.unit[0].get_enthalpy_flow_terms("Sol")
    solid_prop.fs.unit[0].get_energy_density_terms("Sol")

    assert_units_consistent(solid_prop)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"Tests used to test the functionality of the test suite itself."

import pytest


# pytestmark = pytest.mark.skip("Tests for the test suite are always skipped under normal circumstances")


@pytest.mark.xfail
def test_function_without_any_required_mark():
    assert True


@pytest.mark.xfail
@pytest.mark.unit
@pytest.mark.component
def test_function_with_too_many_required_marks():
    assert True
